import pickle
import time
import logging
from os import makedirs
from os.path import dirname, exists, isabs, join
from typing import Any, Optional

from privex.helpers.cache.CacheAdapter import CacheAdapter
from privex.helpers.exceptions import CacheNotFound
from privex.helpers.common import empty, empty_if, is_true
from privex.helpers import settings


log = logging.getLogger(__name__)


def _cache_result_expired(res, _auto_purge=True) -> bool:
    if empty(res): return False
    if empty(res.expires_at, zero=True) or not _auto_purge: return False
    return float(res.expires_at) <= time.time()


class SqliteCache(CacheAdapter):
    """
    An SQLite3 backed implementation of :class:`.CacheAdapter`. Creates and uses a semi-global Sqlite instance via
    :py:mod:`privex.helpers.plugin` by default.

    To allow for a wide variety of Python objects to be safely stored and retrieved from Sqlite, this class
    uses the :py:mod:`pickle` module for serialising + un-serialising values to/from Sqlite.

    **Basic Usage**::

        >>> from privex.helpers.cache import SqliteCache
        >>> rc = SqliteCache()
        >>> rc.set('hello', 'world')
        >>> rc['hello']
        'world'


    **Disabling Pickling**

    In some cases, you may need interoperable caching with other languages. The :py:mod:`pickle` serialisation
    technique is extremely specific to Python and is largely unsupported outside of Python. Thus if you need
    to share Sqlite cache data with applications in other languages, then you must disable pickling.

    **WARNING:** If you disable pickling, then you must perform your own serialisation + de-serialization on
    complex objects such as ``dict``, ``list``, ``Decimal``, or arbitrary classes/functions after getting
    or setting cache keys.

    **Disabling Pickle per instance**

    Pass ``use_pickle=False`` to the constructor, or access the attribute directly to disable pickling for a
    single instance of SqliteCache (not globally)::

        >>> rc = SqliteCache(use_pickle=False)  # Opt 1. Disable pickle in constructor
        >>> rc.use_pickle = False               # Opt 2. Disable pickle on an existing instance


    **Disabling Pickle by default on any new instances**

    Change the static attribute :py:attr:`.pickle_default` to ``False`` to disable the use of pickle by default
    across any new instances of SqliteCache::

        >>> SqliteCache.pickle_default = False



    """
    
    pickle_default: bool = True
    """
    Change this to ``False`` to disable the use of :py:mod:`pickle` by default for any new instances
    of this class.
    """
    
    use_pickle: bool
    """If ``True``, will use :py:mod:`pickle` for serializing objects before inserting into Sqlite, and
    un-serialising objects retrieved from Sqlite3. This attribute is set in :py:meth:`.__init__`.

    Change this to ``False`` to disable the use of :py:mod:`pickle` - instead values will be passed to / returned
    from Sqlite3 as-is, with no serialisation (this may require you to manually serialize complex types such
    as ``dict`` and ``Decimal`` before insertion, and un-serialise after retrieval).
    """
    
    last_purged_expired: Optional[int] = None
    
    def __init__(self, db_file: str = None, memory_persist=False, use_pickle: bool = None, connection_kwargs: dict = None, *args, **kwargs):
        """
        :class:`.SqliteCache` uses an auto-generated database filename / path by default, based on the name of the currently running
        script ( retrieved from ``sys.argv[0]`` ), allowing for persistent caching - without any manual configuration of the adapter,
        nor the requirement for any running background services such as ``redis`` / ``memcached``.
        
        

        :param str db_file:     (Optional) Name of / path to Sqlite3 database file to create/use for the cache.
        
        :param bool memory_persist: Use a shared in-memory database, which can be accessed by other instances of this class (in this
                                    process) - which is cleared after all memory connections are closed.
                                    Shortcut for ``db_file='file::memory:?cache=shared'``

        :param bool use_pickle: (Default: ``True``) Use the built-in ``pickle`` to serialise values before
                                storing in Sqlite3, and un-serialise when loading from Sqlite3
        
        :param dict connection_kwargs: (Optional) Additional / overriding kwargs to pass to :meth:`sqlite3.connect` when
                                        :class:`.SqliteCacheManager` initialises it's sqlite3 connection.
        
        :keyword int purge_every: (Default: 300) Expired + abandoned cache records are purged using the DB manager method
                                  :meth:`.SqliteCacheManager.purge_expired` during :meth:`.get` / :meth:`.set` calls. To avoid
                                  performance issues, the actual :meth:`.SqliteCacheManager.purge_expired` method is only called
                                  if at least ``purge_every`` seconds have passed since the last purge was
                                  triggered ( :attr:`.last_purged_expired` )

        """
        from privex.helpers.cache.post_deps import SqliteCacheManager
        super().__init__(*args, **kwargs)
        self.db_file: str = empty_if(db_file, SqliteCacheManager.DEFAULT_DB)
        self.db_folder = None
        if ':memory:' not in self.db_file:
            if not isabs(self.db_file): self.db_file = join(SqliteCacheManager.DEFAULT_DB_FOLDER, self.db_file)
            self.db_folder = dirname(self.db_file)
            if not exists(self.db_folder):
                log.debug("Folder for database doesn't exist. Creating: %s", self.db_folder)
                makedirs(self.db_folder)
        self.connection_kwargs = empty_if(connection_kwargs, {}, itr=True)
        self.memory_persist = is_true(memory_persist)
        self._wrapper = None
        self.purge_every = kwargs.get('purge_every', 300)
        self.use_pickle = self.pickle_default if use_pickle is None else use_pickle
    
    @property
    def purge_due(self) -> bool:
        lpe = SqliteCache.last_purged_expired
        return empty(lpe, zero=True) or (time.time() - lpe) >= self.purge_every
    
    @property
    def wrapper(self):
        if not self._wrapper:
            from privex.helpers.cache.post_deps import SqliteCacheManager
            self._wrapper = SqliteCacheManager(
                db=self.db_file, connection_kwargs=self.connection_kwargs, memory_persist=self.memory_persist
            )
        return self._wrapper
    
    def purge_expired(self, force=False) -> Optional[int]:
        if self.purge_due or force:
            log.debug("%s - Expired items purge is due (or force is True). Purging expired cache items...", self.__class__.__name__)
            res = self.wrapper.purge_expired()
            log.debug("Finished purging expired items. Total expired cache items deleted: %s", res)
            SqliteCache.last_purged_expired = time.time()
            return res
        return None
    
    def get(self, key: str, default: Any = None, fail: bool = False, _auto_purge=True) -> Any:
        key = str(key)
        _not_found_msg = f'Cache key "{key}" was not found.'
        if _auto_purge: self.purge_expired()
        res = self.wrapper.find_cache_key(key)
        if _cache_result_expired(res, _auto_purge=_auto_purge):
            log.debug("Caller attempted to retrieve expired key '%s', but _auto_purge is True - auto-removing expired key %s", key, key)
            self.remove(key)
            res = None
            _not_found_msg += ' (key was expired - auto-removed)'
        if empty(res):
            if fail: raise CacheNotFound(_not_found_msg)
            return default
        return pickle.loads(res.value) if self.use_pickle else res.value
    
    def set(self, key: str, value: Any, timeout: Optional[int] = settings.DEFAULT_CACHE_TIMEOUT, _auto_purge=True):
        if _auto_purge: self.purge_expired()
        v = pickle.dumps(value) if self.use_pickle else value
        return self.wrapper.set_cache_key(str(key), v, expires_secs=timeout)
    
    def remove(self, *key: str) -> bool:
        removed = 0
        for k in key:
            k = str(k)
            try:
                self.get(k, fail=True, _auto_purge=False)
                self.wrapper.delete_cache_key(k)
                removed += 1
            except CacheNotFound:
                pass
        
        return removed == len(key)
    
    def update_timeout(self, key: str, timeout: int = settings.DEFAULT_CACHE_TIMEOUT) -> Any:
        key, timeout = str(key), int(timeout)
        v = self.get(key=key, fail=True, _auto_purge=False)
        self.set(key=key, value=v, timeout=timeout, _auto_purge=False)
        return v

    def connect(self, db=None, *args, connection_kwargs=None, memory_persist=None, **kwargs):
        c_kwargs = dict(
            connection_kwargs=empty_if(connection_kwargs, self.connection_kwargs),
            memory_persist=empty_if(memory_persist, self.memory_persist)
        )
        c_kwargs = {**c_kwargs, **kwargs}
        from privex.helpers.cache.post_deps import SqliteCacheManager
        self._wrapper = SqliteCacheManager(empty_if(db, self.db_file), *args, **c_kwargs)
        return self._wrapper

    # noinspection PyProtectedMember
    def close(self):
        cls_name = self.__class__.__name__
        if self._wrapper is not None:
            log.debug("Closing Synchronous Sqlite3 instance %s._wrapper", cls_name)
            if self._wrapper._conn:
                try:
                    self._wrapper._conn.close()
                except Exception:
                    log.exception("Unexpected error while closing %s._wrapper._conn", cls_name)
            self._wrapper._conn = None
            self._wrapper = None



"""
Functions, classes and/or types which either **are**, or are related to Python
variable storage types (``dict``, ``tuple``, ``list``, ``set`` etc.)

Object-like Dictionaries (dict's)
---------------------------------

Have you ever wanted a dictionary that works like an object, where you can get/set dictionary keys using 
attributes (``x.something``) as easily as you can with items (``x['something']``)?

We did. So we invented :class:`.DictObject`, a sub-class of the built-in :class:`dict`, making it compatible
with most functions/methods which expect a :class:`dict` (e.g. :meth:`json.dumps`).

You can create a new :class:`.DictObject` and use it just like a ``dict``, or you can convert an existing
``dict`` into a ``DictObject`` much like you'd cast any other builtin type.

It can also easily be cast back into a standard ``dict`` when needed, without losing any data.

Creating a new DictObject and using it
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Since :class:`.DictObject` is a subclass of the builtin :class:`dict`, you can instantiate a new
:class:`.DictObject` in the same way you would use the standard :class:`dict` class::

        >>> d = DictObject(hello='world')
        >>> d
        {'hello': 'world'}
        >>> d['hello']
        'world'
        >>> d.hello
        'world'
        >>> d.lorem = 'ipsum'
        >>> d['orange'] = 'banana'
        >>> d
        {'hello': 'world', 'lorem': 'ipsum', 'orange': 'banana'}


Converting an existing dictionary (dict) into a DictObject
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

You can convert an existing ``dict`` into a :class:`.DictObject` in the same way you'd convert
any other object into a ``dict``::

    >>> y = {"hello": "world", "example": 123}
    >>> x = DictObject(y)
    >>> x.example
    123
    >>> x['hello']
    'world'
    >>> x.hello = 'replaced'
    >>> x
    {'hello': 'replaced', 'example': 123}

It also works vice versa, you can convert a :class:`.DictObject` instance back into a :class:`dict` just as
easily as you converted the `dict` into a `DictObject`.

    >>> z = dict(x)
    >>> z
    {'hello': 'replaced', 'example': 123}



Dict-able NamedTuple's
----------------------

While :func:`collections.namedtuple`'s can be useful, they have some quirks, such as not being able to access
fields by item/key (``x['something']``). They also expose a method ``._asdict()``, but cannot be directly casted
into a :class:`dict` using ``dict(x)``.

Our :func:`.dictable_namedtuple` collection is designed to fix these quirks.

What is a dictable_namedtuple and why use it?
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Unlike the normal :func:`.namedtuple` types, ``dictable_namedtuple``s add extra convenience functionality:

    * Can access fields via item/key: ``john['first_name']``

    * Can convert instance into a dict simply by casting: ``dict(john)``

    * Can set new items/attributes on an instance, even if they weren't previously defined.
    
    * NOTE: You cannot edit an original namedtuple field defined on the type, those remain read only

There are three functions available for working with ``dictable_namedtuple`` classes/instances,
each for different purposes.

  * :py:func:`.dictable_namedtuple` - Create a new ``dictable_namedtuple`` type for instantiation.

  * :py:func:`.convert_dictable_namedtuple` - Convert an existing **namedtuple instance** (not a type/class) into
    a ``dictable_namedtuple`` instance.
  
  * :py:func:`.subclass_dictable_namedtuple` - Convert an existing **namedtuple type/class** (not an instance) into
    a ``dictable_namedtuple`` type for instantiation.


Importing dictable_namedtuple functions
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. code-block:: python

    from collections import namedtuple
    from privex.helpers import dictable_namedtuple, convert_dictable_namedtuple, subclass_dictable_namedtuple

Creating a NEW dictable_namedtuple type and instance
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you're **creating a new Named Tuple**, and you want it to support dictionary-like access, and
have it able to be converted into a dict simply through ``dict(my_namedtuple)``, then you want
:py:func:`.dictable_namedtuple`

.. code-block:: python

    Person = dictable_namedtuple('Person', 'first_name last_name')
    john = Person('John', 'Doe')
    dave = Person(first_name='Dave', last_name='Smith')
    print(dave['first_name'])       # Prints:  Dave
    print(dave.first_name)          # Prints:  Dave
    print(john[1])                  # Prints:  Doe
    print(dict(john))               # Prints:  {'first_name': 'John', 'last_name': 'Doe'}

Converting an existing namedtuple instance into a dictable_namedtuple instance
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


If you have **existing Named Tuple instances**, e.g. returned from a python library, then you can use
:py:func:`.convert_dictable_namedtuple` to convert them into ``dictable_namedtuple``'s and gain all the
functionality mentioned at the start of this section. 

.. code-block:: python

    Person = namedtuple('Person', 'first_name last_name')  # This is an existing namedtuple "type" or "class"
    john = Person('John', 'Doe')  # This is an existing namedtuple instance
    john.first_name               # This works on a standard namedtuple. Returns: John
    john[1]                       # This works on a standard namedtuple. Returns: Doe
    john['first_name']            # However, this would throw a TypeError.
    dict(john)                    # And this would throw a ValueError.

    # We can now convert 'john' into a dictable_namedtuple, which will retain the functionality of a
    # namedtuple, but add to the functionality by allowing dict-like key access, updating/creating new
    # fields, as well as painlessly casting to a dictionary.

    d_john = convert_dictable_namedtuple(john)
    d_john.first_name               # Returns: John
    d_john[1]                       # Returns: Doe
    d_john['first_name']            # Returns: 'John'
    dict(d_john)                    # Returns: {'first_name': 'John', 'last_name': 'Doe'}

Converting an existing namedtuple type/class into a dictable_namedtuple type/class
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you have **existing Named Tuple type/class** then you can use :py:func:`.subclass_dictable_namedtuple`
to convert the type/class into a ``dictable_namedtuple`` type/class and gain all the functionality mentioned 
at the start of this section. (**NOTE:** it's usually easier to just replace your ``namedtuple`` calls 
with ``dictable_namedtuple``)

.. code-block:: python

    Person = namedtuple('Person', 'first_name last_name')  # This is an existing namedtuple "type" or "class"
    # We can now convert the 'Person' type into a dictable_namedtuple type.
    d_Person = subclass_dictable_namedtuple(Person)
    # Then we can use this converted type to create instances of Person with dictable_namedtuple functionality.
    john = d_Person('John', 'Doe')
    john.first_name               # Returns: John
    john[1]                       # Returns: Doe
    john['first_name']            # Returns: 'John'
    dict(john)                    # Returns: {'first_name': 'John', 'last_name': 'Doe'}



**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import copy
import functools
import inspect
import os
import sys
import types
from os.path import dirname, abspath
from collections import namedtuple, OrderedDict
from json import JSONDecodeError
from types import MemberDescriptorType
from typing import Any, Callable, Dict, Optional, NamedTuple, Union, Type, List, Generator, Iterable, TypeVar

# from privex.helpers.decorators import mock_decorator

from privex.helpers.types import AUTO, T, K
import logging
import warnings

log = logging.getLogger(__name__)


def _mock_decorator(*dec_args, **dec_kwargs):
    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)
        return wrapper
    return _decorator


def generate_class(
        name: str, qualname: str = None, module: str = None, bases: Union[tuple, list] = None,
        attributes: Dict[str, Any] = None, **kwargs
) -> Any:
    """
    A small helper function for dynamically generating classes / types.

    **Basic usage**

    Generating a simple class, with an instance constructor, a basic instance method, and an instance factory classmethod::

        >>> import random
        >>> from privex.helpers.collections import generate_class
        >>> def hello_init(self, example: int):
        ...     self.example = example
        ...
        >>> Hello = generate_class(
        ...     'Hello', module='hello',
        ...     attributes=dict(
        ...         __init__=hello_init, lorem=lambda self: self.example * 10,
        ...         make_hello=classmethod(lambda cls: cls(random.randint(1, 100)))
        ...     )
        ... )
        ...
        >>> h = Hello(123)
        >>> h.lorem()
        1230
        >>> j = Hello.make_hello()
        >>> j.example
        77
        >>> j.lorem()
        770

    Generating a child class which inherits from an existing class (the parent(s) can also be a generated classes)::

        >>> World = generate_class(
        ...     'World', module='hello', bases=(Hello,), attributes=dict(ipsum=lambda self: float(self.example) / 3)
        ... )
        >>> w = World(130)
        >>> w.lorem()
        1300
        >>> w.ipsum()
        43.333333333333336

    :param str name:         The name of the class, e.g. ``Hello``
    :param str qualname:     (Optional) The qualified name of the class, e.g. for nested classes ``A -> B -> C``, class ``C``
                             would have the ``__name__``: ``C`` and ``__qualname__``: ``A.B.C``
    :param str module:       (Optional) The module the class should appear to belong to (sets ``__module__``)
    :param tuple|list bases: (Optional) A tuple or list of "base" / "parent" classes for inheritance.
    :param dict attributes:  (Optional) A dictionary of attributes to add to the class. (can include constructor + methods)
    :param kwargs:
    :return:
    """
    qualname = name if qualname is None else qualname
    bases = (object,) if bases is None else (tuple(bases) if not isinstance(bases, tuple) else bases)
    attributes = {} if attributes is None else attributes
    attributes['__module__'] = attributes.get('__module__', module)
    # kwargs = dict(kwargs)
    
    x = type(name, bases, attributes)
    x.__name__ = name
    x.__qualname__ = qualname
    x.__module__ = module
    
    return x


def generate_class_kw(name: str, qualname: str = None, module: str = None, bases: Union[tuple, list] = None, **kwargs) -> Type:
    """
    Same as :func:`.generate_class`, but instead of a :class:`dict` ``attributes`` parameter - all additional keyword arguments
    will be used for ``attributes``

    **Example**::

        >>> def lorem_init(self, ipsum=None):
        ...     self._ipsum = ipsum
        ...
        >>> Lorem = generate_class_kw('Lorem',
        ...     __init__=lorem_init, hello=staticmethod(lambda: 'world'),
        ...     ipsum=property(lambda self: 0 if self._ipsum is None else self._ipsum)
        ... )
        >>> l = Lorem()
        >>> l.ipsum()
        0
        >>> l.hello()
        'world'

    """
    return generate_class(name, qualname, module, bases, attributes=dict(kwargs))


def copy_func(f: Callable, rewrap_classmethod=True, name=None, qualname=None, module=AUTO, **kwargs) -> Union[Callable, classmethod]:
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
    if isinstance(f, classmethod):
        fn = copy_func(f.__func__)
        return classmethod(fn) if rewrap_classmethod else fn
    g = types.FunctionType(f.__code__, f.__globals__, name=name if name is not None else f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    g.__qualname__ = g.__name__ if qualname is None else qualname
    g.__module__ = getattr(f, '__module__') if module is AUTO else module
    return g


def _q_copy(obj: K, key: str = None, deep_private: bool = False, quiet: bool = False, fail: bool = False, **kwargs) -> K:
    should_copy = kwargs.pop('should_copy', True)
    if not should_copy:
        log.debug("Not copying object '%s' (key '%s') as should_copy is False", repr(obj), key)
        return obj
    # By default, deep_private is false, which means we avoid deep copying any attributes which have keys starting with __
    # This is because they're usually objects/types we simply can't deepcopy without issues.
    if not deep_private and key is not None and key.startswith('__'):
        log.debug("Not deep copying key '%s' as deep_private is true and key begins with __", key)
        return obj
    use_copy_func = kwargs.get('use_copy_func', True)
    try:
        if use_copy_func and any([inspect.isfunction(obj), inspect.ismethod(obj), isinstance(obj, classmethod)]):
            copied = copy_func(obj)
        else:
            copied = copy.deepcopy(obj)
        return copied
    except Exception as ex:
        if fail:
            raise ex
        log_args = "Exception while deep copying object %s ( %s ) - using normal ref. Ex: %s %s", key, obj, type(ex), str(ex)
        if quiet:
            log.debug(*log_args)
        else:
            log.warning(*log_args)
    return obj


COPY_CLASS_BLACKLIST = [
    '__dict__', '__slots__', '__weakref__'
]


def _copy_class_dict(obj: Type[T], name, deep_copy=True, deep_private=False, **kwargs) -> Union[Type[T], type]:
    """
    Internal function used by :func:`.copy_class`

    Make a deep copy of the :class:`.type` / class ``obj`` (for standard classes, which use ``__dict__``)
    """
    orig_dict, filt_dict = dict(obj.__dict__), {}
    attr_blacklist = kwargs.get('blacklist', COPY_CLASS_BLACKLIST)
    # Try to deep copy each attribute's value from the original __dict__ in 'orig_dict' into 'filt_dict'.
    # Attributes that are private (start with '__') will use standard references by default (if deep_private is False),
    # along with any attributes that fail to be deep copied.
    for k, v in orig_dict.items():
        # if not deep_copy: break
        if k in attr_blacklist: continue
        filt_dict[k] = v if not deep_copy else _q_copy(v, key=k, deep_private=deep_private, **kwargs)
    
    bases = kwargs.get('bases', obj.__bases__ if kwargs.get('use_bases', True) else (object,))
    return type(name, bases, filt_dict if deep_copy else orig_dict)


def _copy_class_slotted(obj: Type[T], name, deep_copy=True, deep_private=False, **kwargs) -> Union[Type[T], type]:
    """
    Internal function used by :func:`.copy_class`

    Make a deep copy of the :class:`.type` / class ``obj`` (for slotted classes, i.e. those which use ``__slots__``)

    Based on a StackOverflow answer by user ``nkpro``: https://stackoverflow.com/a/61823543/2648583
    """
    slots = obj.__slots__ if type(obj.__slots__) != str else (obj.__slots__,)
    orig_dict, slotted_members = {}, {}
    attr_blacklist = kwargs.get('blacklist', COPY_CLASS_BLACKLIST)

    for k, v in obj.__dict__.items():
        if k in attr_blacklist: continue
        dcval = v if not deep_copy else _q_copy(v, k, deep_private=deep_private, **kwargs)
        if k not in slots:
            orig_dict[k] = dcval
        elif type(v) != MemberDescriptorType:
            slotted_members[k] = dcval
    
    bases = kwargs.get('bases', obj.__bases__ if kwargs.get('use_bases', True) else (object,))
    new_obj = type(name, bases, orig_dict)
    for k, v in slotted_members.items():
        setattr(new_obj, k, v)
    
    return new_obj


DEFAULT_ALLOWED_DUPE = [
    '__annotations__', '__doc__', '__init__', '__getattr__', '__setattr__', '__len__', '__sizeof__',
    '__getattribute__', '__getitem__', '__setitem__', '__delitem__', '__str__', '__repr__',
    '__del__', '__delattr__', '__enter__', '__exit__', '__aenter__', '__aexit__',
    '__next__', '__iter__', '__hash__', '__call__', '__dir__', '__get__', '__set__', '__contains__'
    '__add__', '__sub__', '__mul__', '__floordiv__', '__div__', '__mod__', '__pow__',
    '__eq__', '__ne__', '__lt__', '__gt__', '__le__', '__ge__', '__cmp__',
    '__copy__', '__deepcopy__', '__getstate__', '__setstate__', '__new__',
]


#
# class Asdf:
#     def __

def copy_class_simple(obj: Type[T], name=None, qualname=None, module=AUTO, allow_attrs: list = None, ban_attrs: list = None, **kwargs):
    """
    This is an alternative to :func:`.copy_class` which simply creates a blank class, then iterates over ``obj.__dict__``,
    using :func:`setattr` to copy each attribute over to the cloned class.
    
    It uses :func:`._q_copy` to safely deep copy any attributes which are object references, and thus need their reference pointers
    severed, to avoid edits to the copy affecting the original (and vice versa).
    
    
    :param obj:          The class to duplicate
    :param str name:     The class name to set on the duplicate. If left as ``None``, the duplicate will retain the original ``obj`` name.
    :param str qualname:           The qualified class name to set on the duplicate.
    :param Optional[str] module:   The module path to set on the duplicate, e.g. ``privex.helpers.common``
    :param list allow_attrs: Optionally, you may specify additional private attributes (ones which start with ``__``) that
                             are allowed to be copied from the original class to the duplicated class.
    :param list ban_attrs:   Optionally, you may blacklist certain attributes from being copied from the original class to the duplicate.
    
                             Blacklisted attributes take priority over whitelisted attributes, so you may use this to cancel out
                             any attributes in the default attribute whitelist :attr:`.DEFAULT_ALLOWED_DUPE` which you don't want
                             to be copied to the duplicated class.
    :param kwargs:
    :keyword tuple|list bases: If specified, overrides the default inherited classes (``obj.__bases__``) which would be
                               set on the duplicated class's ``__bases__``.
    :return:
    """
    kwargs = dict(kwargs)

    name = name if name is not None else obj.__name__
    qualname = qualname if qualname is not None else name
    module = module if module is not AUTO else obj.__module__
    
    mkr = generate_class(
        name, qualname, module, bases=kwargs.pop('bases', getattr(obj, '__bases__')), attributes=kwargs.pop('dict_attrs', None)
    )
    allow_attrs = [] if allow_attrs is None else allow_attrs
    ban_attrs = [] if ban_attrs is None else ban_attrs
    allow_attrs += [d for d in DEFAULT_ALLOWED_DUPE if d not in allow_attrs]
    
    for k, v in obj.__dict__.items():
        if k in ban_attrs or (k.startswith('__') and k not in allow_attrs): continue
        try:
            setattr(mkr, k, _q_copy(v, **kwargs))
        except Exception:
            log.exception(f"Failed to copy attribute {k} (value: {v}) to duplicated {obj.__name__} class.")
    # if kwargs.get('bases') is not None:
    #     bases = kwargs.pop('bases')
    #     mkr.__bases__ = bases if isinstance(bases, tuple) else tuple(bases)
    if kwargs.get('str_func') is not None:
        str_func = kwargs.pop('str_func')
        mkr.__str__ = (lambda self: str_func) if isinstance(str_func, str) else str_func

    # mkr.__name__ = name if name is not None else obj.__name__
    # mkr.__qualname__ = qualname if qualname is not None else mkr.__name__
    # mkr.__module__ = module if module is not None else obj.__module__
    return mkr


def copy_class(obj: Type[T], name=None, deep_copy=True, deep_private=False, **kwargs) -> Union[Type[T], type]:
    """
    Attempts to create a full copy of a :class:`.type` or class, severing most object pointers such as attributes containing a
    :class:`.dict` / :class:`.list`, along with classes or instances of classes.

    Example::

        >>> class SomeClass:
        >>>     example = 'lorem ipsum'
        >>>     data = ['hello', 'world']
        >>>     testing = 123
        >>>
        >>> from privex.helpers import copy_class
        >>> OtherClass = copy_class(SomeClass, name='OtherClass')

    If you then append to the :class:`.list` attribute ``data`` on both SomeClass and OtherClass - with a different item
    appended to each class, you'll see that the added item was only added to ``data`` for that class, and not to the other class,
    proving the original and the copy are independent from each other::

        >>> SomeClass.data.append('lorem')
        >>> OtherClass.data.append('ipsum')
        >>> SomeClass.data
        ['hello', 'world', 'lorem']
        >>> OtherClass.data
        ['hello', 'world', 'ipsum']


    :param Type[T] obj: A :class:`.type` / class to attempt to duplicate, deep copying each individual object in the class, to
                        avoid any object pointers shared between the original and the copy.
    :param str|None name: The class name to use for the copy of ``obj``. If not specified, defaults to the original class name from ``obj``
    :param bool deep_copy: (Default: ``True``) If True, uses :func:`copy.deepcopy` to deep copy each attribute in ``obj`` to the copy.
                            If False, then standard references will be used, which may result in object pointers being copied.
    :param bool deep_private: (Default: ``False``) If True, :func:`copy.deepcopy` will be used on "private" class attributes,
                              i.e. ones that start with ``__``. If False, attributes starting with ``__`` will not be deep copied,
                              only a standard assignment/reference will be used.
    :param kwargs: Additional advanced settings (see ``keyword`` pydoc entries for this function)

    :keyword bool use_bases: (Default: ``True``) If True, copy the inheritance (bases) from ``obj`` into the class copy.
    :keyword bool quiet: (Default ``False``) If True, log deep copy errors as ``debug`` level (usually silent in production apps)
                         instead of the louder ``warning``.
    :keyword tuple bases: A :class:`.tuple` of classes to use as "bases" (inheritance) for the class copy. If not specified,
                          copies ``__bases__`` from the original class.
    :keyword str module: If specified, overrides the module ``__module__`` in the class copy with this string, instead of copying from
                         the original class.
    :return Type[T] obj_copy: A deep copy of the original ``obj``
    """
    # If no class name was passed as an attribute, then we copy the name from the original class.
    if not name:
        name = obj.__name__
    
    # Depending on whether 'obj' is a normal class using __dict__, or a slotted class using __slots__, we need to handle the
    # deep copying of the class differently.
    if hasattr(obj, '__slots__'):
        new_obj = _copy_class_slotted(obj, name=name, deep_copy=deep_copy, deep_private=deep_private, **kwargs)
    else:
        new_obj = _copy_class_dict(obj, name=name, deep_copy=deep_copy, deep_private=deep_private, **kwargs)
    
    # Override the module path string if the user specified 'module' as a kwarg
    module = kwargs.get('module')
    if module is not None:
        new_obj.__module__ = module
    return new_obj


def _create_mocker_copy(name=None, **kwargs) -> Union[type, Type["Mocker"]]:
    kwargs['dict_attrs'] = kwargs.get('dict_attrs', {})
    
    def n(self): return self.__class__.__name__
    def m(self): return self.__class__.__module__
    
    nm = copy_func(n, name='__name__', qualname=f"{name}.__name__", module=kwargs.get('module', AUTO))
    # mm = copy_func(m, name='__module__', qualname=f"{name}.__module__", module=kwargs.get('module', AUTO))
    kwargs['dict_attrs']['__name__'] = property(nm)
    # kwargs['dict_attrs']['__module__'] = property(mm)
    
    return copy_class_simple(Mocker, name=name, **kwargs)


class Mocker(object):
    """
    This mock class is designed to be used either to act as a stand-in "noop" (no operation) object, which
    could be used either as a drop-in replacement for a failed module / class import, or for certain unit tests.

    If you need additional functionality such as methods having actual behaviour, you can set attributes on a
    Mocker instance to either a lambda, or point them at a real function/method::

        >>> m = Mocker()
        >>> m.some_func = lambda a: a+1
        >>> m.some_func(5)
        6


    **Example use case - fallback for unimportant module imports**

    Below is a real world example of using :class:`.Mocker` and :py:func:`privex.helpers.decorators.mock_decorator`
    to simulate :py:mod:`pytest` - allowing your tests to run under the standard :py:mod:`unittest` framework if
    a user doesn't have pytest (as long as your tests aren't critically dependent on PyTest).

    Try importing ``pytest`` then fallback to a mock pytest::

        >>> try:
        ...     import pytest
        ... except ImportError:
        ...     from privex.helpers import Mocker, mock_decorator
        ...     print('Failed to import pytest. Using privex.helpers.Mocker to fake pytest.')
        ...     # Make pytest pretend to be the class 'module' (the class actually used for modules)
        ...     pytest = Mocker.make_mock_class('module')
        ...     # To make pytest.mark.skip work, we add the fake module 'mark', then set skip to `mock_decorator`
        ...     pytest.add_mock_module('mark')
        ...     pytest.mark.skip = mock_decorator
        ...

    Since we added the mock module ``mark``, and set the attribute ``skip`` to point at ``mock_decorator``, the
    test function ``test_something`` won't cause a syntax error. ``mock_decorator`` will just call test_something()
    which doesn't do anything anyway::

        >>> @pytest.mark.skip(reason="this test doesn't actually do anything...")
        ... def test_something():
        ...     pass
        >>>
        >>> def test_other_thing():
        ...     if True:
        ...         return pytest.skip('cannot test test_other_thing because of an error')
        ...
        >>>

    **Generating "disguised" mock classes**

    If you need the mock class to appear to have a certain class name and/or module path, you can generate
    "disguised" mock classes using :py:meth:`.make_mock_class` like so:

        >>> redis = Mocker.make_mock_class('Redis', module='redis')
        >>> redis
        <redis.Redis object at 0x7fd7402ea4a8>

    **A :class:`.Mocker` instance has the following behaviour**

    * Attributes that don't exist result in a function being returned, which accepts any arguments / keyword args,
      and simply returns ``None``

    Example::

        >>> m = Mocker()
        >>> repr(m.randomattr('hello', world=123))
        'None'


    * Arbitrary attributes ``x.something`` and items ``x['something']`` can be set on an instance, and they will
      be similarly returned when they're accessed. Attributes and items share the same key/value's, so the
      following examples are all accessing the same data::

    Example::

        >>> m = Mocker()
        >>> m.example = 'hello'
        >>> m['example'] = 'world'
        >>> print(m.example)
        world
        >>> print(m['example'])
        world

    * You can add arbitrary "modules" to a Mocker instance. With only the ``name`` argument, :py:meth:`.add_mock_module`
      will add a "module" under the instance, which is really just another :class:`.Mocker` instance.

    Example::

        >>> m = Mocker()
        >>> m.add_mock_module('my_module')
        >>> m.my_module.example = 'hello'
        >>> print(m.my_module['example'], m.my_module.example)
        hello hello


    """
    # _ALLOWED_DUPE = [
    #     '__module__', '__annotations__', '__doc__', '__init__', '__getattr__', '__setattr__', '__getitem__',
    #     '__setitem__'
    # ]
    mock_modules: dict
    mock_attrs: dict
    
    def __init__(self, modules: dict = None, attributes: dict = None, *args, **kwargs):
        self.mock_attrs = {} if attributes is None else attributes
        self.mock_modules = {} if modules is None else modules
    
    @classmethod
    def make_mock_module(cls, mod_name: str, attributes: dict = None, modules: dict = None, built_in=False, **kwargs):
        mod_base = kwargs.pop('module_base', _module_dir() if built_in else dirname(dirname(dirname(abspath(__file__)))))
        mod_file = os.path.join(mod_base, kwargs.pop('module_file', os.path.join(*mod_name.split('.'), '__init__.py')))
        fix_funcs = kwargs.get('fix_funcs', True)
        
        attributes = {} if not attributes else attributes
        
        if fix_funcs:
            for k, v in attributes.items():
                if inspect.isfunction(v) or inspect.ismethod(v):
                    v.__module__ = mod_name
        modrep = f"<module '{mod_name}' from '{mod_file}'>"
        def x(self): return modrep
        
        # copy_func(x, name='__str__', qualname='__str__', module=mod_name)
        attributes['__repr__'] = attributes.get('__repr__', copy_func(x, name='__repr__', qualname='__repr__', module=mod_name))
        attributes['__str__'] = attributes.get('__str__', copy_func(x, name='__str__', qualname='__str__', module=mod_name))
        attributes['__file__'] = attributes.get('__file__', mod_file)
        return cls.make_mock_class('module', attributes=attributes, modules=modules, **kwargs)

    @classmethod
    def make_mock_class(
        cls, name='Mocker', instance=True, simple=False, attributes: dict = None, modules: dict = None, **kwargs
    ) -> Union[Any, "Mocker", Type["Mocker"]]:
        """
        Return a customized mock class or create an instance which appears to be named ``name``

        Allows code which might check ``x.__class__.__name__`` to believe it's the correct object.

        Using the kwarg ``module`` you can change the module that the class / instance appears to have been imported
        from, allowing for quite deceiving fake classes and instances.

        **Example usage**::

            >>> redis = Mocker.make_mock_class('Redis', module='redis')
            >>> # As seen below, the class appears to be called Redis, and even claims to be from the module `redis`
            >>> redis
            <redis.Redis object at 0x7fd7402ea4a8>
            >>> print(f'Module: {redis.__module__} - Class Name: {redis.__class__.__name__}')
            Module: redis - Class Name: Redis

        **Creating methods/attributes dynamically**

        You can set arbitrary attributes to point at a function, or just set them to a lambda::

            >>> redis.exists = lambda key: 1
            >>> redis.exists('hello')
            1
            >>> redis.hello()  # Non-existent attributes just act as a function that eats any args and returns None
            None


        :param name: The name to write onto the mock class's ``__name__`` (and ``__qualname__`` if not specified)
        :param bool instance: If ``True`` then the disguised mock class will be returned as an instance. Otherwise
                              the raw class itself will be returned for you to instantiate yourself.
        :param bool simple:   When ``True``, generates a very basic, new class - not based on :class:`.Mocker`, which contains
                              the attributes/methods defined in the param ``attributes``.
        :param kwargs: All kwargs (other than ``qualname``) are forwarded to ``__init__`` of the disguised class
                       if ``instance`` is True.
        :param dict attributes: If ``simple`` is True, then this dictionary of attributes is used to generate the class's
                                attributes, methods, and/or constructor.
                                If ``simple`` is False, and ``instance`` is True, these attributes are passed to the constructor
                                of the :class:`.Mocker` clone that was generated.
        :param dict modules: If ``simple`` is False, and ``instance`` is True, this dict of modules are passed to the constructor
                                of the :class:`.Mocker` clone that was generated.
        
        :key str qualname: Optionally specify the "qualified name" to insert into ``__qualname__``. If this isn't
                           specified, then ``name`` is used for qualname, which is fine for most cases anyway.
        :key str module: Optionally override the module namespace that the class is supposedly from. If not specified,
                         then the class will just inherit this module (``privex.helpers.common``)
        :return:
        """
        qualname = kwargs.pop('qualname', name)
        mod_name = kwargs.pop('module', __name__)
        attributes = {} if attributes is None else attributes
        modules = {} if modules is None else modules
        if simple:
            c = generate_class(
                name, qualname=qualname, module=mod_name, bases=kwargs.pop('bases', None), attributes=attributes
            )
            return c(**kwargs) if instance else c
        c = _create_mocker_copy(name, deep_private=True, module=mod_name, qualname=qualname)

        # OuterMocker.__name__ = name
        
        # if mod_name is not None:
        #     OuterMocker.__module__ = kwargs['module']
        str_func = attributes.pop('__str__', None)
        repr_func = attributes.pop('__repr__', None)
        if str_func is not None: c.__str__ = str_func
        if repr_func is not None: c.__repr__ = repr_func
        return c(modules=modules, attributes=attributes, **kwargs) if instance else c
        
    def add_mock_module(self, name: str, value=None, mock_attrs: dict = None, mock_modules: dict = None):
        """
        Add a fake sub-module to this Mocker instance.

        Example::

            >>> m = Mocker()
            >>> m.add_mock_module('my_module')
            >>> m.my_module.example = 'hello'
            >>> print(m.my_module['example'], m.my_module.example)
            hello hello


        :param str name: The name of the module to add.
        :param value: Set the "module" to this object, instead of an instance of :class:`.Mocker`
        :param dict mock_attrs: If ``value`` is ``None``, then this can optionally contain a dictionary of
                                attributes/items to pre-set on the Mocker instance.
        :param dict mock_modules: If ``value`` is ``None``, then this can optionally contain a dictionary of
                                 "modules" to pre-set on the Mocker instance.
        """
        mock_attrs = {} if mock_attrs is None else mock_attrs
        mock_modules = {} if mock_modules is None else mock_modules
        
        self.mock_modules[name] = Mocker(modules=mock_modules, attributes=mock_attrs) if value is None else value
    
    def add_mock_modules(self, *module_list, _dict_to_attrs=True, _parse_dict=True, **module_map):
        """
        
            >>> hello = Mocker.make_mock_class('Hello')
            >>> hello.add_mock_modules(
            ...     world={
            ...         'lorem': 'ipsum',
            ...         'dolor': 123,
            ...     }
            ... )
        
        :param module_list:
        :param _parse_dict:
        :param _dict_to_attrs:
        :param module_map:
        :return:
        """
        module_map = dict(module_map)
        for m in module_list:
            log.debug("Adding simple mock module from module_list: %s", m)
            self.add_mock_module(m)
        for k, v in module_map.items():
            m_val, m_attrs, m_modules = v, {}, {}
            if isinstance(v, dict):
                if _parse_dict:
                    _m_val = None
                    if 'value' in v: _m_val = v['value']
                    if 'attrs' in v:
                        log.debug("Popping 'attrs' from kwarg '%s' value as attributes for module: %s", k, v)
                        m_attrs = {**m_attrs, **v.pop('attrs')}
                    if 'modules' in v:
                        log.debug("Popping 'modules' from kwarg '%s' value as attributes for module: %s", k, v)
                        m_modules = {**m_modules, **v.pop('modules')}
                    if not _dict_to_attrs or not all([_m_val is None, m_attrs is None, m_modules is None]):
                        log.debug("Setting module value to value of kwarg '%s': %s", k, v)
                        m_val = _m_val if m_attrs is None and m_modules is None else None
                if _dict_to_attrs:
                    log.debug("Importing kwarg '%s' value as attributes for module: %s", k, v)
                    m_attrs = {**m_attrs, **v}
                    
            self.add_mock_module(k, m_val, mock_attrs=m_attrs, mock_modules=m_modules)
    
    @classmethod
    def _duplicate_cls(cls, name=None, qualname=None, module=None, **kwargs) -> Type["Mocker"]:
        return _create_mocker_copy(name=name, qualname=qualname, module=module, **kwargs)
    
    def _duplicate_ins(self, name=None, qualname=None, module=None, **kwargs) -> "Mocker":
        mkr = _create_mocker_copy(name=name, qualname=qualname, module=module, **kwargs)
        return mkr(modules=self.mock_modules, attributes=self.mock_attrs)
    
    def __getattr__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            pass
        try:
            if item in object.__getattribute__(self, 'mock_modules'):
                return self.mock_modules[item]
        except AttributeError:
            pass
        try:
            if item in object.__getattribute__(self, 'mock_attrs'):
                return self.mock_attrs[item]
        except AttributeError:
            pass
        
        return lambda *args, **kwargs: None
    
    def __setattr__(self, key, value):
        if key in ['mock_attrs', 'mock_modules']:
            return object.__setattr__(self, key, value)
        m = object.__getattribute__(self, 'mock_attrs')
        m[key] = value
    
    def __getitem__(self, item):
        try:
            return self.__getattr__(item)
        except AttributeError as ex:
            raise KeyError(str(ex))
    
    def __setitem__(self, key, value):
        try:
            self.__setattr__(key, value)
        except AttributeError as ex:
            raise KeyError(str(ex))
    
    @property
    def __name__(self):
        return self.__class__.__name__

    def __dir__(self) -> Iterable[str]:
        base_attrs = list(object.__dir__(self))
        extra_attrs = list(self.mock_attrs.keys()) + list(self.mock_modules.keys())
        return base_attrs + extra_attrs


def _module_dir():
    import collections
    col_dir = dirname(abspath(collections.__file__))
    return dirname(col_dir)
    

dataclasses_mock = Mocker.make_mock_module(
    'dataclasses',
    attributes=dict(
        dataclass=_mock_decorator,
        asdict=lambda obj, dict_factory=dict: dict_factory(obj),
        astuple=lambda obj, tuple_factory=tuple: tuple_factory(obj),
        is_dataclass=lambda obj: False,
        field=lambda *args, **kwargs: kwargs.get('default', kwargs.get('default_factory', lambda: None)()),
    ), built_in=True
)
"""
This is a :class:`.Mocker` instance which somewhat emulates the Python 3.7+ :mod:`dataclasses` module,
including the :func:`dataclasses.dataclass` decorator.
"""

try:
    # noinspection PyCompatibility
    import dataclasses
    # noinspection PyCompatibility
    from dataclasses import dataclass, field
except (ImportError, ImportWarning, AttributeError, KeyError) as e:
    warnings.warn(
        f"Failed to import dataclasses (added in Python 3.7). Setting placeholders for typing. "
        f"If you're running Python older than 3.7 and want to use the dataclass features in privex-helpers, "
        f"please update to Python 3.7+, or run 'pip3 install -U dataclasses' to install the backported dataclass emulation "
        f"library for older Python versions.", category=ImportWarning
    )
    # To avoid a severe syntax error caused by the missing dataclass types, we generate a dummy dataclasses module, along with a
    # dummy dataclass and field class so that type annotations such as Type[dataclass] don't cause the module to throw a syntax error.
    # noinspection PyTypeHints
    dataclasses = dataclasses_mock
    # noinspection PyTypeHints
    dataclass = dataclasses.dataclass
    field = dataclasses.field


class DictObject(dict):
    """
    A very simple :class:`dict` wrapper, which allows you to read and write dictionary keys using attributes
    (dot notation) PLUS standard item (key / square bracket notation) access.

    **Example Usage (creating and using a new DictObject)**::

        >>> d = DictObject(hello='world')
        >>> d
        {'hello': 'world'}
        >>> d['hello']
        'world'
        >>> d.hello
        'world'
        >>> d.lorem = 'ipsum'
        >>> d['orange'] = 'banana'
        >>> d
        {'hello': 'world', 'lorem': 'ipsum', 'orange': 'banana'}

    **Example Usage (converting an existing dict)**::

        >>> y = {"hello": "world", "example": 123}
        >>> x = DictObject(y)
        >>> x.example
        123
        >>> x['hello']
        'world'
        >>> x.hello = 'replaced'
        >>> x
        {'hello': 'replaced', 'example': 123}

    """

    def __getattr__(self, item):
        """When an attribute is requested, e.g. ``x.something``, forward it to ``dict['something']``"""
        if hasattr(super(), item):
            return super().__getattribute__(item)
        try:
            return super().__getitem__(item)
        except KeyError as ex:
            raise AttributeError(str(ex))

    def __setattr__(self, key, value):
        """When an attribute is set, e.g. ``x.something = 'abcd'``, forward it to ``dict['something'] = 'abcd'``"""
        if hasattr(super(), key):
            return super().__setattr__(key, value)
        try:
            return super().__setitem__(key, value)
        except KeyError as ex:
            raise AttributeError(str(ex))
    
    def __dir__(self) -> Iterable[str]:
        return list(dict.__dir__(self)) + list(self.keys())
        

class OrderedDictObject(OrderedDict):
    """
    Ordered version of :class:`.DictObject` - dictionary with attribute access.
    See :class:`.DictObject`
    """
    def __getattr__(self, item):
        """When an attribute is requested, e.g. ``x.something``, forward it to ``dict['something']``"""
        if hasattr(super(), item):
            return super().__getattribute__(item)
        try:
            return super().__getitem__(item)
        except KeyError as ex:
            raise AttributeError(str(ex))
    
    def __setattr__(self, key, value):
        """When an attribute is set, e.g. ``x.something = 'abcd'``, forward it to ``dict['something'] = 'abcd'``"""
        if hasattr(super(), key):
            return super().__setattr__(key, value)
        try:
            return super().__setitem__(key, value)
        except KeyError as ex:
            raise AttributeError(str(ex))

    def __dir__(self) -> Iterable[str]:
        return list(OrderedDict.__dir__(self)) + list(self.keys())


class MockDictObj(DictObject):
    """
    This is a masqueraded :class:`.DictObject` made to look like the builtin :class:`dict` by
    editing the class name, qualname and module.

    It may improve compatibility when passing :class:`.DictObject` to certain third-party 
    functions/methods.

    Note: this isn't enough to fool a ``type(x) is dict`` check.
    """


MockDictObj.__name__ = 'dict'
MockDictObj.__qualname__ = 'dict'
MockDictObj.__module__ = 'builtins'



def is_namedtuple(*objs) -> bool:
    """
    Takes one or more objects as positional arguments, and returns ``True`` if ALL passed objects
    are namedtuple instances

    **Example usage**

    First, create or obtain one or more NamedTuple objects::

        >>> from collections import namedtuple
        
        >>> Point, Person = namedtuple('Point', 'x y'), namedtuple('Person', 'first_name last_name')

        >>> pt1, pt2 = Point(1.0, 5.0), Point(2.5, 1.5)
        >>> john = Person('John', 'Doe')
    
    We'll also create a ``tuple``, ``dict``, and ``str`` to show they're detected as invalid::

        >>> normal_tuple, tst_dict, tst_str = (1, 2, 3,), dict(hello='world'), "hello world"
    
    First we'll call :func:`.is_namedtuple` with our Person NamedTuple object ``john``::

        >>> is_namedtuple(john)
        True
    
    As expected, the function shows ``john`` is in-fact a named tuple.

    Now let's try it with our two Point named tuple's ``pt1`` and ``pt2``, plus our Person named tuple ``john``.

        >>> is_namedtuple(pt1, john, pt2)
        True
    
    Since all three arguments were named tuples (even though pt1/pt2 and john are different types), the function
    returns ``True``.

    Now we'll test with a few objects that clearly aren't named tuple's::

        >>> is_namedtuple(tst_str)   # Strings aren't named tuples.
        False
        >>> is_namedtuple(normal_tuple)    # A plain bracket tuple is not a named tuple.
        False
        >>> is_namedtuple(john, tst_dict)  # ``john`` is a named tuple, but a dict isn't, thus False is returned.
        False


    Original source: https://stackoverflow.com/a/2166841

    :param Any objs: The objects (as positional args) to check whether they are a NamedTuple
    :return bool is_namedtuple: ``True`` if all passed ``objs`` are named tuples.
    """
    if len(objs) == 0: raise AttributeError("is_namedtuple expects at least one argument")

    for x in objs:
        t = type(x)
        b = t.__bases__
        if tuple not in b: return False
        f = getattr(t, '_fields', None)
        if not isinstance(f, tuple): return False
        if not all(type(n) == str for n in f): return False
    return True


def convert_dictable_namedtuple(nt_instance, typename=None, module=None, **kwargs) -> Union[NamedTuple, Dict]:
    """
    Convert an existing :func:`collections.namedtuple` instance into a dictable_namedtuple instance.

    **Example**

    First we create a namedtuple type ``Person``

        >>> from collections import namedtuple
        >>> Person = namedtuple('Person', 'first_name last_name')
    
    Next we create an instance of ``Person`` called John Doe, and we can confirm it's a normal namedtuple, as we
    can't access first_name by item/key.

        >>> john = Person('John', 'Doe')
        >>> john['first_name']
        TypeError: tuple indices must be integers or slices, not str
    
    Using :func:`.convert_dictable_namedtuple`, we can convert ``john`` from a normal ``namedtuple``, into
    a ``dictable_namedtuple``.
    
    This enables many convenience features (see :func:`.dictable_namedtuple` for more info) 
    such as easy casting to a :class:`dict`, and accessing fields by item/key (square brackets)::

        >>> from privex.helpers import convert_dictable_namedtuple
        >>> d_john = convert_dictable_namedtuple(john)
        >>> d_john
        Person(first_name='John', last_name='Doe')
        >>> d_john['first_name']
        'John'
        >>> dict(d_john)
        {'first_name': 'John', 'last_name': 'Doe'}
    

    :param nt_instance: An instantiated namedtuple object (using a type returned from :func:`collections.namedtuple`)
    :param str typename: Optionally, you can change the name of your instance's class, e.g. if you provide a ``Person`` 
                         instance, but you set this to ``Man``, then this will return a ``Man`` instance, like so:
                         ``Man(first_name='John', last_name='Doe')``
    :param str module: Optionally, you can change the module that the type class belongs to. Otherwise it will inherit the module path
                       from the class of your instance.
    :key bool read_only: (Default: ``False``) If set to ``True``, the outputted dictable_namedtuple instance will not
                         allow new fields to be created via attribute / item setting.
    :return dictable_namedtuple: The instance you passed ``nt_instance``, converted into a dictable_namedtuple
    """

    nt_class = nt_instance.__class__
    module = nt_class.__module__ if module is None else module

    dnt_class = subclass_dictable_namedtuple(nt_class, typename=typename, module=module, **kwargs)
    return dnt_class(**nt_instance._asdict())


def subclass_dictable_namedtuple(named_type: type, typename=None, module=None, **kwargs) -> type:
    """
    Convert an existing :func:`collections.namedtuple` **type** into a dictable_namedtuple.

    If you have an INSTANCE of a type (e.g. it has data attached), use :func:`.convert_dictable_namedtuple`

    **Example**::

        >>> from collections import namedtuple
        >>> from privex.helpers import subclass_dictable_namedtuple
        >>> # Create a namedtuple type called 'Person'
        >>> orig_Person = namedtuple('Person', 'first_name last_name')
        >>> # Convert the 'Person' type into a dictable_namedtuple
        >>> Person = subclass_dictable_namedtuple(orig_Person)
        >>> john = Person('John', 'Doe')   # Create an instance of this dictable_namedtuple Person
        >>> john['middle_name'] = 'Davis'
    
    :param type named_type: A NamedTuple type returned from :func:`collections.namedtuple`
    :param str typename: Optionally, you can change the name of your type, e.g. if you provide a ``Person`` 
                         class type, but you set this to ``Man``, then this will return a ``Man`` class type.
    :param str module: Optionally, you can change the module that the type class belongs to. Otherwise it will
                       inherit the module path from ``named_type``.
    :key bool read_only: (Default: ``False``) If set to ``True``, the outputted dictable_namedtuple type will not allow
                         new fields to be created via attribute / item setting.
    :return type dictable_namedtuple: Your ``named_type`` converted into a dictable_namedtuple type class.
    """
    typename = named_type.__name__ if typename is None else typename
    module = named_type.__module__ if module is None else module
    read_only = kwargs.pop('read_only', False)

    _dt = make_dict_tuple(typename, ' '.join(named_type._fields), read_only=read_only)

    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass

    if module is not None:
        _dt.__module__ = module
    return _dt


def make_dict_tuple(typename, field_names, *args, **kwargs):
    """
    Generates a :func:`collections.namedtuple` type, with added / modified methods injected to make it
    into a ``dictable_namedtuple``.
    
    Note: You probably want to be using :func:`.dictable_namedtuple` instead of calling this directly.
    """
    read_only = kwargs.pop('read_only', False)
    module = kwargs.pop('module', None)
    
    # Create a namedtuple type to use as a base
    BaseNT = namedtuple(typename, field_names, **kwargs)

    def __init__(self, *args, **kwargs):
        self.__dict__['_extra_items'] = dict()
        for i, a in enumerate(list(args)):
            self.__dict__[self._fields[i]] = a
        for k, a in kwargs.items():
            self.__dict__[k] = a
    
    def __iter__(self):
        """This ``__iter__`` method allows for casting a dictable_namedtuple instance using ``dict(my_nt)``"""
        for k in self._fields: yield (k, getattr(self, k),)

    def __getitem__(self, item):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt['abc']`` or ``my_nt[0]``"""
        if type(item) is int:
            return self.__dict__[self._fields[item]]
        return getattr(self, item)

    def __getattr__(self, item):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt.abcd``"""
        try:
            _v = object.__getattribute__(self, '_extra_items')
            return _v[item]
        except (KeyError, AttributeError):
            return object.__getattribute__(self, item)

    def __setitem__(self, key, value):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt['abc'] = 'def'``"""
        if hasattr(self, key):
            return tuple.__setattr__(self, key, value)
        if self._READ_ONLY:
            raise KeyError(f"{self.__class__.__name__} is read only. You cannot set a non-existent field.")
        self._extra_items[key] = value
        if key not in self._fields:
            tuple.__setattr__(self, '_fields', self._fields + (key,))

    def __setattr__(self, key, value):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt.abcd = 'def'``"""
        if key in ['_extra_items', '_fields'] or key in self._fields:
            return tuple.__setattr__(self, key, value)
        if self._READ_ONLY:
            raise AttributeError(f"{self.__class__.__name__} is read only. You cannot set a non-existent field.")
        self._extra_items[key] = value
        if key not in self._fields:
            tuple.__setattr__(self, '_fields', self._fields + (key,))

    def _asdict(self):
        """
        The original namedtuple ``_asdict`` doesn't work with our :meth:`.__iter__`, so we override it
        for compatibility. Simply calls ``return dict(self)`` to convert the instance to a dict.
        """
        return dict(self)

    def __repr__(self):
        _n = ', '.join(f"{name}='{getattr(self, name)}'" for name in self._fields)
        return f"{self.__class__.__name__}({_n})"

    # Inject our methods defined above into the namedtuple type BaseNT
    BaseNT.__getattr__ = __getattr__
    BaseNT.__getitem__ = __getitem__
    BaseNT.__setitem__ = __setitem__
    BaseNT.__setattr__ = __setattr__
    BaseNT._asdict = _asdict
    BaseNT.__repr__ = __repr__
    BaseNT.__iter__ = __iter__
    BaseNT.__init__ = __init__
    BaseNT._READ_ONLY = read_only
    
    # Create a class for BaseNT with tuple + object mixins, allowing things like __dict__ to function properly
    # and allowing for tuple.__setattr__ / object.__getattribute__ calls.
    class K(BaseNT, tuple, object):
        pass

    # Get the calling module so we can overwrite the module name of the class.
    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass
    
    # Overwrite the type name + module to match the originally requested typename
    K.__name__ = BaseNT.__name__
    K.__qualname__ = BaseNT.__qualname__
    K.__module__ = module

    return K


def dictable_namedtuple(typename, field_names, *args, **kwargs) -> Union[Type[namedtuple], dict]:
    """
    Creates a dictable_namedtuple type for instantiation (same usage as :func:`collections.namedtuple`) - unlike
    namedtuple, dictable_namedtuple instances allow item (dict-like) field access, support writing and can be
    painlessly converted into dictionaries via ``dict(my_namedtuple)``.

    Named tuple instances created from ``dictable_namedtuple`` types are generally backwards compatible with
    any code that expects a standard :func:`collections.namedtuple` type instance.

    **Quickstart**

        >>> from privex.helpers import dictable_namedtuple
        >>> # Define a dictable_namedtuple type of 'Person', which has two fields - first_name and last_name
        >>> p = dictable_namedtuple('Person', 'first_name last_name')
        >>> john = p('John', 'Doe')    # Alternatively you can do p(first_name='John', last_name='Doe')
        >>> john.first_name            # You can retrieve keys either via attributes (dot notation)
        'John'
        >>> john['last_name']          # Via named keys (square brackets)
        'Doe'
        >>> john[1]                    # Or, via indexed keys (square brackets, with integer keys)
        'Doe'
        >>> john.middle_name = 'Davis' # You can also update / set new keys via attribute/key/index
        >>> dict(john)                 # Newly created keys will show up as normal in dict(your_object)
        {'first_name': 'John', 'last_name': 'Doe', 'middle_name': 'Davis'}
        >>> john                       # As well as in the representation in the REPL or when str() is called.
        Person(first_name='John', last_name='Doe', middle_name='Davis')

    This function adds / overrides the following methods on the generated namedtuple type:

    * _asdict
    * __iter__
    * __getitem__
    * __getattribute__
    * __setitem__
    * __setattr__
    * __repr__

    Extra functionality compared to the standard :func:`.namedtuple` generated classes:

        * Can access fields via item/key: ``john['first_name']``

        * Can convert instance into a dict simply by casting: ``dict(john)``

        * Can set new items/attributes on an instance, even if they weren't previously defined.
          ``john['middle_name'] = 'Davis'`` or ``john.middle_name = 'Davis'``

    **Example Usage**

    First we'll create a named tuple typle called ``Person``, which takes two arguments, first_name and last_name.

        >>> from privex.helpers import dictable_namedtuple
        >>> Person = dictable_namedtuple('Person', 'first_name last_name')
    
    Now we'll create an instance of ``Person`` called ``john``. These instances look like normal ``namedtuple``'s, and
    should be generally compatible with any functions/methods which deal with named tuple's.

        >>> john = Person('John', 'Doe')   # Alternatively you can do Person(first_name='John', last_name='Doe')
        >>> john
        Person(first_name='John', last_name='Doe')
    
    Unlike a normal ``namedtuple`` type instance, we can access fields by attribute (``.first_name``), index (``[0]``), 
    AND by item/key name (``['last_name']``).

        >>> john.first_name
        'John'
        >>> john[0]
        'John'
        >>> john['last_name']
        'Doe'
    
    Another potentially useful feature, is that you can also update / create new fields, via your preferred method
    of field notation (other than numbered indexes, since those don't include a field name)::

        >>> john['middle_name'] = 'Davis'
        >>> john.middle_name = 'Davis'

    We can also convert ``john`` into a standard dictionary, with a simple ``dict(john)`` cast. You can see that
    the new field we added (``middle_name``) is present in the dictionary serialized format.

        >>> dict(john)
        {'first_name': 'John', 'last_name': 'Doe', 'middle_name': 'Davis'}
    

    :param str typename: The name used for the namedtuple type/class
    :param str field_names: One or more field names separated by spaces, e.g. ``'id first_name last_name address'``
    :key bool read_only: (Default: ``False``) If set to ``True``, the outputted dictable_namedtuple instance will not
                         allow new fields to be created via attribute / item setting.
    :return Type[namedtuple] dict_namedtuple: A dict_namedtuple type/class which can be instantiated with the given
                             ``field_names`` via positional or keyword args.
    """
    module = kwargs.get('module', None)
    read_only = kwargs.pop('read_only', False)
    
    # As per namedtuple's comment block, we need to set __module__ to the frame
    # where the named tuple is created, otherwise it can't be pickled properly.
    # This also ensures that __module__ would match that of a normal namedtuple()
    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass

    return make_dict_tuple(typename, field_names, module=module, read_only=read_only, *args, **kwargs)


class Dictable:
    """
    A small abstract class for use with Python 3.7 dataclasses.

    Allows dataclasses to be converted into a ``dict`` using the standard ``dict()`` function:

        >>> @dataclass
        >>> class SomeData(Dictable):
        ...     a: str
        ...     b: int
        ...
        >>> mydata = SomeData(a='test', b=2)
        >>> dict(mydata)
        {'a': 'test', 'b': 2}

    Also allows creating dataclasses from arbitrary dictionaries, while ignoring any extraneous dict keys.

    If you create a dataclass using a ``dict`` and you have keys in your ``dict`` that don't exist in the dataclass,
    it'll generally throw an error due to non-existent kwargs:

        >>> mydict = dict(a='test', b=2, c='hello')
        >>> sd = SomeData(**mydict)
        TypeError: __init__() got an unexpected keyword argument 'c'

    Using ``from_dict`` you can simply trim off any extraneous dict keys:

        >>> sd = SomeData.from_dict(**mydict)
        >>> sd.a, sd.b
        ('test', 2)
        >>> sd.c
        AttributeError: 'SomeData' object has no attribute 'c'



    """

    def __iter__(self):
        # Allow casting into dict()
        for k, v in self.__dict__.items(): yield (k, v,)

    def __getitem__(self, key):
        """
        When the instance is accessed like a dict, try returning the matching attribute.
        """
        if hasattr(self, key): return getattr(self, key)
        raise KeyError(key)

    def __setitem__(self, key, value):
        """
        When the instance is modified like a dict, try setting the matching attribute.
        """
        return setattr(self, key, value)

    def get(self: Union[Type["dataclasses.dataclass"], "dataclasses.dataclass"], key, fallback=None):
        names = set([f.name for f in dataclasses.fields(self)])
        return getattr(self, key) if key in names else fallback

    @classmethod
    def from_dict(cls, env):
        # noinspection PyArgumentList
        return cls(**{
            k: v for k, v in env.items()
            if k in inspect.signature(cls).parameters
        })

    @classmethod
    def from_list(cls: Type[T], obj_list: Iterable[dict]) -> Generator[T, None, None]:
        for o in obj_list:
            yield cls.from_dict(o)


class DictDataClass(Dictable):
    """
    This is a base class for use with Python 3.7+ :class:`.dataclass` 's, designed to make dataclasses more interoperable with
    existing dictionaries, and allows them to be used like dictionaries, similar to :class:`.Dictable`, but more powerful /
    flexible.
    
    The most notable difference between this and :class:`.Dictable` - is that DictDataClass uses the attribute :attr:`.raw_data` on your
    dataclass to store any excess attributes when your dataclass is initialised from a dictionary with :meth:`.from_dict`
    or :meth:`.from_list`, allowing you to retrieve any dictionary keys which couldn't be stored on your dataclass instance.
    
    Basic Example::
    
        >>> from dataclasses import dataclass, field
        >>> from privex.helpers import DictDataClass, DictObject
        >>> from typing import Union
        >>>
        >>> @dataclass
        >>> class ExampleDataclass(DictDataClass):
        ...     example: str = 'hello world'
        ...     lorem: int = 999
        ...     raw_data: Union[dict, DictObject] = field(default_factory=DictObject, repr=False)
        ...     ### ^ The raw, unmodified data that was passed as kwargs, as a dictionary
        ...     ### For DictDataClass to work properly, you must include the raw_data dataclass field in your dataclass.
        ...
        >>> edc = ExampleDataclass.from_dict(dict(example='test', hello='this is an example'))
        >>> edc.example
        'test'
        >>> dict(edc)
        {'example': 'test', 'lorem': 999}
    
    Thanks to :attr:`.raw_data` - you can access any extraneous items contained within the dictionary used in :meth:`.from_dict`
    as if they were part of your dataclass. You can also access and set any attribute using standard dictionary item syntax with square
    brackets like so::
    
        >>> edc.hello               # This is not in the original dataclass attributes, but is proxied from raw_data
        'this is an example'
        >>> edc['hello']            # Also works with item / dict syntax
        'this is an example'
        >>> edc['hello'] = 'world'  # You can set attributes using "key" / dict-like syntax
        >>> edc.hello
        'world'
        >>> edc.hello = 'test'      # You can also set raw_data keys using standard attribute dot-notation syntax.
        >>> edc['hello']
        'test'
     
    **Dictionary casting modes / ``__iter__`` configuration modes**
    
    There are a total of four (4) :class:`.dict`` conversion modes that you may use for a given class.
    
    These modes control whether :attr:`.raw_data` is used when an instance is being casted via ``dict(obj)``, the order in which it's
    merged with the instance attributes, along with the option to include **just** the dataclass attributes, or **just** the raw_data.
    
    Available :attr:`.DictConfig.dict_convert_mode` options::
    
        * **Fallback / Dataclass / Instance Attributes Only** - When ``dict_convert_mode`` is empty (``None`` or ``""``), or
          it can't be matched against a pre-defined conversion mode, when an instance is converted into a :class:`.dict` - only the
          attributes of the instance are used - :attr:`.raw_data` keys are ignored.
          
          ``dict_convert_mode`` settings: ``None``, ``""``, ``"none"`` or any other invalid option.
        
        * Raw Data Only ( ``raw`` / ``raw_data`` ) - When this mode is used, converting the instance to a ``dict`` will
          effectively just return ``raw_data``, while still enforcing the ``dict_exclude`` and ``dict_listify`` settings.
          
          ``dict_convert_mode`` settings: ``"raw"``, ``"raw_data"``, ``"rawdata"``, or any other value beginning with ``raw``
        
        
        * Merge with Dataclass Priority - In this mode, both :attr:`.raw_data` and the instance's attributes are used when converting
          into a :class:`.dict` - first the :attr:`.raw_data` dictionary is taken, and we merge the instance attributes on top of it.
          
          This means the instance/dataclass attributes take priority over the ``raw_data`` attributes, which will generally result
          in only ``raw_data`` keys which don't exist on the instance having their values used in the final ``dict``.
        
          ``dict_convert_mode`` settings: ``"merge_dc"`` / ``"merge_ins"`` / ``"merge_dataclass"``
        
        * Merge with :attr:`.raw_data` Priority - In this mode, both :attr:`.raw_data` and the instance's attributes are used when
          converting into a :class:`.dict` - first the instance attributes are converted into a dict, then we merge the
          :attr:`.raw_data` dictionary on top of it.
          
          This means the :attr:`.raw_data` keys take priority over the instance/dataclass attributes, which will generally result in
          only instance attributes which don't exist in ``raw_data`` having their values used in the final ``dict``.
        
          ``dict_convert_mode`` settings: ``"merge_rd"`` / ``"merge_raw"`` / ``"merge_raw_data"``
    
    By default, the conversion mode is set to ``merge_dc``, which means when an instance is converted into a :class:`.dict` - the
    dataclass instance attributes are merged on top of the raw_data dictionary, meaning raw_data is included, but dataclass attributes
    take priority over :attr:`.raw_data` keys.
    
    **Changing the conversion mode**
    
        >>> from dataclasses import dataclass, field
        >>>
        >>> @dataclass
        >>> class MyDataclass(DictDataClass):
        ...     class DictConfig:
        ...         dict_convert_mode = 'merge_dc'
        ...
        ...     hello: str
        ...     lorem: str = 'ipsum'
        ...     raw_data: Union[dict, DictObject] = field(default_factory=DictObject, repr=False)
        >>> dc = MyDataclass.from_dict(dict(hello='test', example=555, test='testing'))
        >>> dc.hello
        'test'
        >>> dc.hello = 'replaced'
        >>> dict(dc)
        {'hello': 'replaced', 'example': 555, 'test': 'testing', 'lorem': 'ipsum'}
        >>> dc.DictConfig.dict_convert_mode = 'merge_raw'
        {'hello': 'test', 'lorem': 'ipsum', 'example': 555, 'test': 'testing'}
        >>> dc.DictConfig.dict_convert_mode = None
        {'hello': 'replaced', 'lorem': 'ipsum'}
    
    """
    raw_data: Union[dict, DictObject]
    
    class _DictConfig:
        dict_convert_mode: Optional[str] = "merge_dc"
        dict_listify: List[str] = []
        """
        Keys which contain iterable's (list/set etc.) of objects such as those based on :class:`.DictDataClass` / :class:`.Dictable`,
        which should be converted into a list of :class:`.dict`'s using ``dict()``.
        
        Example:
        
            >>> @dataclass
            >>> class Order(DictDataClass):
            ...     hello: str = 'world'
            >>>
            >>> class MyDataclass(DictDataClass):
            ...     class DictConfig:
            ...         dict_listify = ['orders']
            ...     lorem = 'ipsum'
            ...     orders: list = [Order(), Order('test')]
            ...
        """
        listify_cast: Union[Type, callable] = DictObject
        dict_exclude_base: List[str] = ['raw_data', 'DictConfig', '_DictConfig']
        dict_exclude: List[str] = []
        """
        A list of attributes / raw_data keys to exclude when your instance is converted into a :class:`.dict`
        """
    
    @property
    def _dc_dict_config(self) -> Union[_DictConfig, DictObject]:
        """Internal property used to merge the base DictDataClass _DictConfig with the current instance's DictConfig"""
        base = {k: v for k, v in self._DictConfig.__dict__.items() if not k.startswith('__')}
        inst = {k: v for k, v in self.DictConfig.__dict__.items() if not k.startswith('__')}
        return DictObject({**base, **inst})
    
    DictConfig = copy_class(_DictConfig, name='DictConfig', quiet=True)

    def __iter__(self):
        cls_name = self.__class__.__name__
        # The raw_data attribute isn't required, and isn't guaranteed to have been set on a dataclass, so
        # we fallback to a blank DictObject if it's not found.
        dict_rd = DictObject()
        if hasattr(self, 'raw_data'):
            dict_rd = dict(self.raw_data)
        
        # dataclasses.asdict can sometimes freak out when a dataclass contains un-serializable objects such as arbitrary
        # class instances (which may even be excluded already in DictConfig. In such a case, we can fallback to __dict__ :)
        try:
            dict_dc = dataclasses.asdict(self, dict_factory=dict)
        except (TypeError, ValueError, JSONDecodeError) as e:
            log.warning("Dictifying %s using dataclasses.asdict failed (%s %s)... falling back to __dict__", cls_name, type(e), str(e))
            dict_dc = dict(self.__dict__)

        from privex.helpers import empty
        dconf = self._dc_dict_config
        mode = "" if empty(dconf.dict_convert_mode, True, True) else str(dconf.dict_convert_mode).lower()
        # By default, if dict_convert_mode is None, or doesn't match any of the pre-defined modes, then it
        # defaults to dict_dc (the instance's dataclass attributes, converted into a dictionary)
        itr = dict_dc
        if mode in ['merge_dc', 'merge_ins', 'merge_dataclass']:
            itr = {**dict_rd, **dict_dc}
        elif mode in ['merge_rd', 'merge_raw', 'merge_raw_data']:
            itr = {**dict_dc, **dict_rd}
        elif mode.startswith('raw'):
            itr = dict_rd
        
        exclude = dconf.dict_exclude_base + dconf.dict_exclude
        
        for k, v in itr.items():
            # Don't include any dict keys which are listed in dict_exclude/dict_exclude_base
            if k in exclude: continue
            # If this key is listed in dict_listify, then we iterate over the current value, casting
            # each element using the type/callable specified in DictConfig.listify_case
            # (defaults to DictObject, which is backwards compatible with code requiring dict's)
            if k in dconf.dict_listify: v = [dconf.listify_cast(o) for o in v]
            yield (k, v,)
    
    def __getitem__(self, key):
        """
        When the instance is accessed like a dict, try returning the matching attribute.
        If the attribute doesn't exist, or the key is an integer, try and pull it from raw_data
        """
        if hasattr(self, 'raw_data') and type(key) is int: return self.raw_data[key]
        if hasattr(self, key): return getattr(self, key)
        if hasattr(self, 'raw_data') and key in self.raw_data: return self.raw_data[key]
        raise KeyError(key)

    def __setitem__(self, key, value):
        """
        When the instance is modified like a dict, try setting the matching attribute.
        If the key is an int, or exists in raw_data, then set it on raw_data instead of as a dataclass attribute.
        """
        if hasattr(self, 'raw_data') and type(key) is int:
            self.raw_data[key] = value
            return
        
        if hasattr(self, 'raw_data') and key in self.raw_data:
            self.raw_data[key] = value
            return
        
        return setattr(self, key, value)
    
    def __getattr__(self, item):
        # object.__getattribute__()
        def ha(key):
            return key in self.__dict__
        
        if ha('raw_data') and type(item) is int:
            return self.__dict__['raw_data'][item]
        if ha(item):
            return self.__dict__[item]
        if ha('raw_data') and item in self.__dict__['raw_data']:
            return self.__dict__['raw_data'][item]
        raise AttributeError(item)
    
    def __setattr__(self, key, value):
        def ha(k): return k in self.__dict__
        
        if ha(key):
            self.__dict__[key] = value
            return
        
        if ha('raw_data'):
            self.__dict__['raw_data'][key] = value
            return
        
        return super().__setattr__(key, value)
    
    @classmethod
    def from_dict(cls: Type[dataclass], obj):
        names = set([f.name for f in dataclasses.fields(cls)])
        clean = {k: v for k, v in obj.items() if k in names}
        if 'raw_data' in names:
            clean['raw_data'] = DictObject(clean.get('raw_data')) if 'raw_data' in clean else DictObject(obj)
        return cls(**clean)
    

DictDataclass = DictDataClass




import platform
import socket
import ssl
import subprocess
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network, ip_address, ip_network
from typing import List, Optional, Union
from privex.helpers import settings
from privex.helpers.common import empty_if, byteify, is_true, stringify
from privex.helpers.exceptions import NetworkUnreachable
from privex.helpers.types import IP_OR_STR, STRBYTES

__all__ = [
    'ip_is_v4', 'ip_is_v6', 'ping', 'IPV4_ALIASES', 'IPV6_ALIASES', 'ip_ver_to_int', 'ip_ver_to_sock',
    'sock_ver', 'is_ip', 'sock_validate_ip'
]


def ip_is_v4(ip: str) -> bool:
    """
    Determines whether an IP address is IPv4 or not

    :param str ip: An IP address as a string, e.g. 192.168.1.1
    :raises ValueError: When the given IP address ``ip`` is invalid
    :return bool: True if IPv6, False if not (i.e. probably IPv4)
    """
    return type(ip_address(ip)) == IPv4Address


def ip_is_v6(ip: str) -> bool:
    """
    Determines whether an IP address is IPv6 or not

    :param str ip: An IP address as a string, e.g. 192.168.1.1
    :raises ValueError: When the given IP address ``ip`` is invalid
    :return bool: True if IPv6, False if not (i.e. probably IPv4)
    """
    return type(ip_address(ip)) == IPv6Address


def ping(ip: str, timeout: int = 30) -> bool:
    """
    Sends a ping to a given IPv4 / IPv6 address. Tested with IPv4+IPv6 using ``iputils-ping`` on Linux, as well as the
    default IPv4 ``ping`` utility on Mac OSX (Mojave, 10.14.6).
    
    Fully supported when using Linux with the ``iputils-ping`` package. Only IPv4 support on Mac OSX.
    
    **Example Usage**::
    
        >>> from privex.helpers import ping
        >>> if ping('127.0.0.1', 5) and ping('::1', 10):
        ...     print('Both 127.0.0.1 and ::1 are up')
        ... else:
        ...     print('127.0.0.1 or ::1 failed to respond to a ping within the given timeout.')
    
    **Known Incompatibilities**:
    
     * NOT compatible with IPv6 addresses on OSX due to the lack of a timeout argument with ``ping6``
     * NOT compatible with IPv6 addresses when using ``inetutils-ping`` on Linux due to separate ``ping6`` command

    :param str ip: An IP address as a string, e.g. ``192.168.1.1`` or ``2a07:e00::1``
    :param int timeout: (Default: 30) Number of seconds to wait for a response from the ping before timing out
    :raises ValueError: When the given IP address ``ip`` is invalid or ``timeout`` < 1
    :return bool: ``True`` if ping got a response from the given IP, ``False`` if not
    """
    ip_obj = ip_address(ip)   # verify IP is valid (this will throw if it isn't)
    if timeout < 1:
        raise ValueError('timeout value cannot be less than 1 second')
    opts4 = {
        'Linux': ["/bin/ping", "-c1", f"-w{timeout}"],
        'Darwin': ["/sbin/ping", "-c1", f"-t{timeout}"]
    }
    opts6 = {'Linux':  ["/bin/ping", "-c1", f"-w{timeout}"]}
    opts = opts4 if ip_is_v4(ip_obj) else opts6
    if platform.system() not in opts:
        raise NotImplementedError(f"{__name__}.ping is not fully supported on platform '{platform.system()}'...")
    
    with subprocess.Popen(opts[platform.system()] + [ip], stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
        out, err = proc.communicate()
        err = err.decode('utf-8')
        if 'network is unreachable' in err.lower():
            raise NetworkUnreachable(f'Got error from ping: "{err}"')
        
        return 'bytes from {}'.format(ip) in out.decode('utf-8')


IPV4_ALIASES = [4, 'v4', '4', 'ipv4', 'ip4', 'inet', 'inet4', socket.AF_INET, str(socket.AF_INET)]
IPV6_ALIASES = [6, 'v6', '6', 'ipv6', 'ip6', 'inet6', socket.AF_INET6, str(socket.AF_INET6)]


def ip_ver_to_int(ver: Union[str, int]) -> int:
    ver = str(ver).lower()
    if ver in IPV4_ALIASES: return 4
    if ver in IPV6_ALIASES: return 6
    return 0


def sock_ver(version) -> Optional[int]:
    version = empty_if(version, 'any', zero=True, itr=True)
    version = str(version).lower()
    if ip_ver_to_int(version) == 4: return socket.AF_INET
    if ip_ver_to_int(version) == 6: return socket.AF_INET6
    return None


ip_ver_to_sock = sock_ver


def ip_sock_ver(ip_addr) -> Optional[int]:
    a = ip_network(ip_addr, strict=False)
    if isinstance(a, (IPv4Address, IPv4Network)): return socket.AF_INET
    if isinstance(a, (IPv6Address, IPv6Network)): return socket.AF_INET6
    return None


def is_ip(addr: str, version: int = None):
    try:
        res = sock_validate_ip(addr, version=version)
        return res
    except AttributeError as e:
        raise e
    except ValueError:
        return False


def sock_validate_ip(addr: IP_OR_STR, version: int, throw=True) -> Optional[Union[IPv4Address, IPv4Address]]:
    ip = ip_address(addr)
    ver = "v4" if ip_is_v4(ip) else "v6"
    if version == socket.AF_INET and ver != 'v4':
        if not throw: return None
        raise AttributeError(f"Passed address '{addr}' was an IPv6 address, but 'version' requested an IPv4 address.")
    if version == socket.AF_INET6 and ver != 'v6':
        if not throw: return None
        raise AttributeError(f"Passed address '{addr}' was an IPv4 address, but 'version' requested an IPv6 address.")
    return ip


def get_ssl_context(
            verify_cert: bool = False, check_hostname: Optional[bool] = None, verify_mode: Optional[int] = None, **kwargs
        ) -> ssl.SSLContext:
    check_hostname = empty_if(check_hostname, is_true(verify_cert))
    verify_mode = empty_if(verify_mode, ssl.CERT_REQUIRED if verify_cert else ssl.CERT_NONE)
    
    ctx = ssl.create_default_context()
    ctx.check_hostname = check_hostname
    ctx.verify_mode = verify_mode
    return ctx


def generate_http_request(
    url="/", host=None, method="GET", user_agent=settings.DEFAULT_USER_AGENT, extra_data: Union[STRBYTES, List[str]] = None,
    body: STRBYTES = None, **kwargs
) -> bytes:
    method, url = stringify(method), stringify(url)
    http_ver = stringify(kwargs.get('http_ver', '1.0'))
    data = f"{method.upper()} {url} HTTP/{http_ver}\n"
    if host is not None: data += f"Host: {stringify(host)}\n"
    if user_agent is not None: data += f"User-Agent: {stringify(user_agent)}\n"
    data = byteify(data)
    
    if extra_data is not None:
        if isinstance(extra_data, list):
            extra_data = [byteify(x) for x in extra_data]
            data += b"\n".join(extra_data)
        else:
            data += byteify(extra_data)
        if not data.endswith(b"\n"): data += b"\n"
    
    if body is not None:
        data += byteify(body)
    if not data.endswith(b"\n"): data += b"\n"
    if not data.endswith(b"\n\n"): data += b"\n"
    return data



"""
Helper functions / classes for using ``git`` within a python application.

Quickstart with the aliases ``get_current_commit``, ``get_current_tag`` and ``get_current_branch``
--------------------------------------------------------------------------------------------------

To save you time, we pre-instantiate :class:`._AsyncGit` at :attr:`._cwd_git` within the module, using the current working
directory as the repo, allowing you to quickly call some of the most common Git functions without having to instantiate
a ``Git()`` instance::

    >>> from privex.helpers import get_current_tag, get_current_branch, get_current_commit
    >>> get_current_commit()
    '8418c964b35d76bcad984f5102ac605be0ae7b58'
    >>> get_current_branch()
    'master'
    >>> get_current_tag()
    '2.14.0'


Using the Git class ( :class:`._AsyncGit` )
-------------------------------------------

Basic usage of :class:`.Git` ( :class:`._AsyncGit` )
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

For full functionality, you'll want to import ``Git`` and create an instance.

Despite the fact that the Git class methods are all async methods, they can be called from within a synchronous app or context, and
they'll generally work as if they were synchronous functions, thanks to the :func:`.awaitable_class` decorator.


    >>> from privex.helpers import Git
    >>> g = Git(repo='/home/user/projects/some_app')
    >>> g.get_current_commit()
    'e962f66650729e2f66395b45e1600f61d8461378'
    >>> g.add("docs", "README.md")
    >>> g.commit("1.2.3 - Added documentation and README.md")
    >>> g.tag('1.2.3')
    >>> g.get_current_tag()
    '1.2.3'

Using Git commands which don't have a method implemented
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The magic method :meth:`._AsyncGit.__getattr__` allows you to call Git sub-commands which aren't yet implemented as a method, simply
by calling the command's name as a method on the class instance.

If the sub-command contains ``-`` 's (dashes) in it's name, simply replace the dashes with underlines - ``__getattr__`` will
automatically rewrite the underlines (``_``) back into dashes (``-``).

For example, we can run ``git count-objects`` by calling the method ``count_objects``, even though ``count_objects`` isn't
implemented as a method. Just like normal methods, we can call it synchronously from non-async functions/methods, and use async ``await``
when using it within asynchronous code::

    >>> await g.count_objects()
    '2061 objects, 8880 kilobytes'
    >>> g.count_objects('-v')
    'count: 2061\nsize: 8880\nin-pack: 0\npacks: 0\nsize-pack: 0\nprune-packable: 0\ngarbage: 0\nsize-garbage: 0'



"""
from os import getcwd
from os.path import isabs, abspath, join
from typing import Tuple, Optional, Callable, Coroutine, Union, Any, AnyStr

from privex.helpers.common import stringify, empty, empty_if
from privex.helpers import STRBYTES
from privex.helpers.asyncx import call_sys_async, awaitable, awaitable_class
from privex.helpers.exceptions import SysCallError


async def _async_sys(proc, *args, write: STRBYTES = None, **kwargs) -> Tuple[str, str]:
    """Small async wrapper function for :func:`.call_sys_async` to simplify calling ``git``, including detecting git errors"""
    stderr_raise = kwargs.pop('stderr_raise', False)
    strip = kwargs.pop('strip', True)
    out, err = await call_sys_async(proc, *args, write=write, **kwargs)
    out, err = stringify(out, if_none=''), stringify(err, if_none='')
    if strip: out, err = out.strip(), err.strip()
    
    if len(err) > 1 and stderr_raise:
        raise SysCallError(f'Git returned an error: "{err}"')
    
    return out, err


def _repo(repo: str = None) -> str:
    """
    A very simple private helper function which ensures the passed ``repo`` is an absolute directory path. If ``repo`` is blank,
    then simply returns the current working directory. Converts relative paths into absolute paths by joining them to :func:`.getcwd`
    """
    repo = getcwd() if empty(repo) else repo
    return join(getcwd(), repo) if not isabs(repo) else repo


class _AsyncGit:
    """
    Git CLI wrapper class - methods can be used both synchronously and async via ``AsyncGit`` / ``Git`` (aliases, both are the same)
    
    This class uses the :func:`.awaitable_class` decorator to make the class work both synchronously and asynchronously, without needing
    to duplicate code.
    
    The :func:`.awaitable_class` decorator detects whether the methods are being called from a synchronous, or an asynchronous
    function/method.
    
    * If they're being called from an async function, then a coroutine will be returned, and must be ``await``'ed.
    
    * If they're being called from a synchronous function, then it will spin up an event loop, run the method in
      the event loop, then return the result transparently.
    
    
    **Synchronous usage**::
        
        >>> from privex.helpers import Git
        >>> g = Git()   # Git is an alias for AsyncGit, there is no difference between them
        >>>
        >>> def some_func():
        ...     current_commit = g.get_current_commit()
        ...     print(current_commit)
        'ac52b28f551825160785f9ea7e96f86ccc869cc1'
    
    **Asynchronous usage**::
        
        >>> from privex.helpers import Git
        >>> g = Git()    # Git is an alias for AsyncGit, there is no difference between them
        >>>
        >>> async def some_func():
        ...     current_commit = await g.get_current_commit()
        ...     print(current_commit)
        'ac52b28f551825160785f9ea7e96f86ccc869cc1'

    """
    repo: Optional[str]
    default_version: str
    
    def __init__(self, repo: str = None, default_version: str = 'HEAD', **kwargs):
        self.repo = None
        self.repo = _repo(repo)
        self.default_version = default_version
    
    def _repo(self, repo=None) -> str:
        return self.repo if empty(repo) and not empty(self.repo) else _repo(repo)

    async def git(self, *args, repo: str = None, strip=True, stderr=False) -> Union[str, Tuple[str, str]]:
        """
        Wrapper async method for calling ``git`` executable command against a repo - casts args to :class:`str`.
        
        This is intended for use internally by :class:`._AsyncGit` methods, but is under a non-private method name to allow you to call
        Git commands semi-directly if the class methods are getting in the way.
        
        Example::
        
            >>> Git().git('commit', '--amend', '-a', '-m', 'lorem ipsum dolor')
        
        
        :param AnyStr|int|float|object|Any args: Positional command line arguments to pass to ``git``. Most formats are fine, as long
                                                 as they can be casted into a :class:`str`
        
        :param str|None repo: The repository to run the command against. When ``None``, uses the value of :attr:`.repo`
        
        :param bool strip: (Default: ``True``) Whether to run :meth:`str.strip` on the string outputs of the command
        
        :param bool stderr: (Default: ``False``) When ``True``, includes the stderr output in the response like so: ``(stdout, stderr)``
                            When ``False``, simply returns ``stdout`` directly as a :class:`str`
          
        :return str stdout: (when ``stderr=False``) The text outputted to stdout by the command
        :return Tuple[str,str] stdout_err: (when ``stderr=True``) A :class:`tuple` containing ``(stdout: str, stderr: str)``
        """
        repo = self._repo(repo)
        out, err = await _async_sys("git", *[stringify(a) for a in args], cwd=repo, strip=strip)
        
        return (out, err) if stderr else out
    
    _git = git
    
    async def init(self, *args, repo: str = None) -> str:
        """
        Use like ``git init``

        **Example**::

            >>> Git().init()
        
        **Initialise a different repo from the one passed in the constructor**::
        
            >>> Git().init(repo="/home/user/myproject")

        """
        return await self.git("init", *args, repo=repo)

    async def add(self, *args, repo: str = None) -> str:
        """
        Use like ``git add``
        
        Example:
            
            >>> Git().add("docs/", "README.md")
        
        """
        return await self.git("add", *args, repo=repo)

    async def branch(self, *args, repo: str = None) -> str:
        return await self.git("branch", *args, repo=repo)

    async def commit(self, message: Optional[Union[str, bool]], *args, repo: Optional[str] = None) -> str:
        """
        Calls ``git commit`` with the arguments ``-m "message" [args]``, where each positional argument passed after the message
        is passed as a command line argument.
        
        To disable prepending ``-m`` to the arguments, pass ``None`` as the ``message`` and only your specified ``args`` will be
        passed along to ``git commit``.
        
        You may also set ``message`` to ``False`` which will call ``git commit --allow-empty-message [args]`` instead of
        prepending ``-m``.
        
        
        **Basic Usage**
        
        Standard commit with a message, plus the argument '-a'::
        
            >>> g = Git()
            >>> g.commit("added example.txt", "-a")
        
        Pass ``None`` as the message to disable prepending ``-m`` to the git arguments. For example, the following call commits
        the current staged changes, while re-using the commit message/author info/timestamp from the previous
        commit ``9e27d7233ac5bc59bc37c0572a401068fbd5e6be``::
        
            >>> g.commit(None, "-C", "9e27d7233ac5bc59bc37c0572a401068fbd5e6be")
        
        Pass ``False`` as the message for an easy way to call ``git commit --allow-empty-message``, plus any additional
        arguments you might pass::
        
            >>> g.commit(False, '-a')
        
        
        :param str|bool message: The git commit message to commit with. Alternatively ``None`` to remove the default ``-m``, or
                                 ``False`` as a shortcut for ``--allow-empty-message`` instead of ``-m``.
        :param args: Additional CLI arguments to pass to ``git commit``
        :param str repo: (as a kwarg only!) An absolute path to a Git repository to run ``git commit`` within. By default, this is
                         ``None`` which results in the repo passed in the constructor ( :attr:`.repo` ) being used.
        :return str stdout: The string text printed to stdout by ``git commit`` while running the command.
        """
        if message is None:
            return await self.git("commit", *args, repo=repo)
        if message is False:
            return await self.git("commit", "--allow-empty-message", *args, repo=repo)
        return await self.git("commit", "-m", stringify(message), *args, repo=repo)

    async def checkout(self, branch: str, *args, repo: str = None, new: bool = False) -> str:
        args = ['-b', branch] + list(args) if new else [branch] + list(args)
        return await self.git("checkout", *args, repo=repo)
    
    async def status(self, *args, repo: str = None, concise=True) -> str:
        if concise: args = ['-s'] + list(args)
        return await self.git("status", *args, repo=repo, strip=False)

    async def tag(self, *args, repo: str = None) -> str:
        return await self.git("tag", *args, repo=repo)
    
    async def get_current_commit(self, version: str = None, repo: str = None) -> str:
        """
        Get current commit hash. Optionally specify ``version`` to get the current commit hash for a branch or tag.
        
        **Examples**::
        
            >>> g = Git()
            >>> await g.get_current_commit()
            'ac52b28f551825160785f9ea7e96f86ccc869cc1'
            >>> await g.get_current_commit('2.0.0')    # Get the commit hash for the tag 2.0.0
            '598584a447ba63212ac3fe798c01941badf1c194'
        
        
        :param str version:  Optionally specify a branch / tag to get the current commit hash for.
        :param str repo:     Optionally specify a specific local repository path to run ``git`` within.
        :return str commit_hash: The current Git commit hash
        """
        return await self.git("rev-parse", empty_if(version, self.default_version), repo=repo)
    
    async def get_current_branch(self, repo: str = None) -> str:
        """
        Get current active branch/tag.

        **Examples**::

            >>> g = Git()
            >>> await g.get_current_branch()
            'master'
            >>> await g.checkout('testing', new=True)    # Create and checkout the branch 'testing'
            >>> await g.get_current_branch()
            'testing'
        

        :param str repo:     Optionally specify a specific local repository path to run ``git`` within.
        :return str current_branch: The name of the current checked out branch or tag
        """
        return await self.git("rev-parse", "--abbrev-ref", 'HEAD', repo=repo)

    async def get_current_tag(self, version: str = None, repo: str = None) -> str:
        """
        Get the latest tag on this branch - useful for detecting current version of your python application.

        **Examples**::

            >>> g = Git()
            >>> await g.get_current_tag()          # Get the latest tag on the active branch
            '2.5.0'
            >>> await g.get_current_tag('develop') # Get the latest tag on the branch 'develop'
            '2.5.3'

        :param str version:  Optionally specify a branch / tag to get the latest tag for.
        :param str repo:     Optionally specify a specific local repository path to run ``git`` within.
        :return str current_tag: The name of the latest tag on this branch.
        """
        return await self.git("describe", "--abbrev=0", "--tags", empty_if(version, self.default_version), repo=repo)
    
    async def log(self, *args, repo: str = None, concise=True):
        if concise: args = ['--oneline'] + list(args)
        return await self.git("--no-pager", "log", *args, repo=repo)

    def __getattr__(self, item: str) -> Union[Callable[[Any, Any, Any, Any, Any], Coroutine], callable, Any]:
        """
        If an attribute doesn't exist, this method will return a :func:`.awaitable` function that simply calls :meth:`.git` with the
        attribute name as the first argument, followed by any additional positional/keyword arguments.
        
        Since a lot of git commands have ``-`` in their name, but Python doesn't support attributes containing ``-``, non-existent
        attributes will have ``_`` replaced with ``-`` when calling :meth:`.git` with their name.
        
        For example, ``git.count_objects('-v')`` would become ``git count-objects -v``
        
        This allows most unimplemented ``git`` sub-commands to function when called as a method, for example, at the
        current point in time, neither the ``describe`` command, nor ``count-objects`` are implemented as a method, but both can still be
        called synchronously or asynchronously and it will work::
        
            >>> from privex.helpers import Git
            >>> g = Git()
            >>> g.describe('--tags')
            '2.14.0-1-g8418c96'
            >>> await g.describe('--tags')
            '2.14.0-1-g8418c96'
            >>> g.count_objects()
            '2061 objects, 8880 kilobytes'
            >>> await g.count_objects()
            '2061 objects, 8880 kilobytes'
        
        
        """
        try:
            value = super().__getattribute__(item)
            return value
        except AttributeError:
            @awaitable
            def _git(*args, **kwargs):
                return self.git(item.replace('_', '-'), *args, **kwargs)
            return _git


AsyncGit = awaitable_class(_AsyncGit)

Git = AsyncGit
_cwd_git = AsyncGit()
"""
:attr:`._cwd_git` is a pre-instantiated :class:`._AsyncGit` instance which uses the current working directory as the current git repo.

This instance is used by global instance method aliases :attr:`.get_current_commit`, :attr:`.get_current_branch`, :attr:`.get_current_tag`,
to allow developers to easily call methods such as :meth:`._AsyncGit.get_current_tag` without having to instantiate the
:class:`.Git` class.
"""

get_current_commit = _cwd_git.get_current_commit
"""Alias for :meth:`.AsyncGit.get_current_commit` (using pre-instantiated AsyncGit at :attr:`._cwd_git`)"""

get_current_branch = _cwd_git.get_current_branch
"""Alias for :meth:`.AsyncGit.get_current_branch` (using pre-instantiated AsyncGit at :attr:`._cwd_git`)"""

get_current_tag = _cwd_git.get_current_tag
"""Alias for :meth:`.AsyncGit.get_current_tag` (using pre-instantiated AsyncGit at :attr:`._cwd_git`)"""

__all__ = [
    'AsyncGit', 'Git', 'get_current_commit', 'get_current_branch', 'get_current_tag', '_repo',
]



import functools
import threading
import logging
import inspect
from contextlib import contextmanager
from time import sleep
from typing import Any, Callable, ContextManager, Set, Type, Union, Optional

import attr
from privex.helpers.common import DictObject, auto_list, empty, empty_if
from privex.helpers.cache import CacheNotFound, cached
from privex.helpers.decorators import r_cache, FO
from privex.helpers.types import AnyNum, AUTO
from privex.helpers.thread import lock_acquire_timeout
from privex.helpers.black_magic import caller_name

log = logging.getLogger(__name__)

__all__ = [
    'CACHE_MGR', 'NO_LOCK', 'ANY_LCK', 'fake_lock_manager', 'CacheSettings', 'z_cache', 'CacheManagerMixin'
]

CACHE_MGR = Union["CacheManagerMixin", Type["CacheManagerMixin"]]
NO_LOCK = type('NoLock', (), {})
ANY_LCK = Optional[Union[threading.Lock, Type[NO_LOCK]]]


@contextmanager
def fake_lock_manager(lock: threading.Lock = None, *args, **kwargs):
    yield True


@attr.s
class CacheSettings:
    cache_prefix: str = attr.ib('pvx_cmgr')
    """The prefix used for all cache keys generated for your class"""
    cache_key_log_name: str = attr.ib('all_cache_keys')
    """The cache key name component for the key used to store a list of all known cache keys for your class"""
    cache_sep: str = attr.ib(':')
    """Separator character(s) used between identifying components within a cache key"""
    default_cache_key_time: Union[float, int] = attr.ib(300)
    """Default number of seconds to cache the log of known cache keys"""
    default_cache_time: Union[float, int] = attr.ib(300)
    """Default number of seconds to cache any objects"""
    _gen_cache_key: Callable[[Any], str] = attr.ib(None)
    _key_add_prefix: Callable[[Any], str] = attr.ib(None)
    
    def gen_cache_key(self, *args, _auto_cache=True, _comps_start: list = None, _comps_end: list = None, **query) -> str:
        # noinspection PyArgumentList
        return self._gen_cache_key(*args, _auto_cache=_auto_cache, _comps_start=_comps_start, _comps_end=_comps_end, **query)
    
    def key_add_prefix(
            self, key: Union[str, Callable[[Any], str]], auto_prefix: bool = True, _auto_cache=True,
            call_args: list = None, call_kwargs: dict = None, _lock: ANY_LCK = None
    ) -> str:
        # noinspection PyArgumentList
        return self._key_add_prefix(
            key, auto_prefix, _auto_cache=_auto_cache, call_args=call_args, call_kwargs=call_kwargs, _lock=_lock
        )
    
    @classmethod
    def from_class(cls, obj: Union["CacheManagerMixin", Type["CacheManagerMixin"]]) -> "CacheSettings":
        _keys = [
            'cache_prefix', 'cache_key_log_name', 'cache_sep', 'default_cache_time',
            'default_cache_key_time', 'gen_cache_key', 'key_add_prefix'
        ]
        # _priv_keys = ['gen_cache_key', 'key_add_prefix']
        constx = {}
        for k in _keys:
            if hasattr(obj, k):
                constx[k] = getattr(obj, k)
        # for k in _priv_keys:
        #     if hasattr(obj, k):
        #         constx[f"_{k}"] = getattr(obj, k)
        return cls(**constx)


def z_cache(cls: CACHE_MGR = None, cache_key: Union[str, callable] = None, cache_time=AUTO, format_args: list = None,
            format_opt: FO = FO.POS_AUTO, extract_class=True, **opts):
    """
    A special method caching decorator which is designed to integrate with classes that extend :class:`.CacheManagerMixin`
    
    This is simply a wrapper for :func:`.r_cache` - it transparently caches the output of a wrapped method/function, but
    unlike :func:`.r_cache`, it's designed to automatically integrate with classes which extend :class:`.CacheManagerMixin` ,
    allowing it to automatically retrieve cache settings such as the cache prefix, default cache time, along with directly calling
    various classmethods which enable logging of newly created ``cache_key`` 's for painless cleanup of cache keys when needed,
    without having to manually track them, or doing the nuclear option of erasing the entire cache system's database.
    
    It integrates with your class by intercepting the first argument of any method calls that are wrapped with :func:`.z_cache`.
    Both standard instance methods and :class:`.classmethod` 's are supported transparently - while static methods will only
    integrate properly with your own class if you set ``cls`` (first argument) on the decorator to point to the class or instance
    containing your cache manager settings and inherited methods from :class:`.CacheManagerMixin`
    
    After you've extended :class:`.CacheManagerMixin`, you can enable argument-based caching for any method, simply by adding
    the decorator line ``@z_cache()`` above it. :func:`.z_cache` will automatically handle the cache key name, including extracting
    your method's passed arguments and inserting them into the cache key, ensuring method calls with certain arguments, are cached
    separately from other method calls with a different set of arguments.
    
    Example::
        
        >>> from privex.helpers.cache.extras import CacheManagerMixin, z_cache
        >>> from time import sleep
        >>>
        >>> class MyClass(CacheManagerMixin):
        ...     # It's recommended, but not necessary - to re-create cache_key_lock with None, as it will ensure your
        ...     # class has a separate threading.Lock() instance from other CacheManagerMixin based classes
        ...     cache_key_lock: Optional[threading.Lock] = None
        ...     # You should override cache_prefix to ensure cache keys auto-created for your class won't
        ...     # conflict with cache keys created by other classes
        ...     cache_prefix: str = 'my_class'
        ...     # While it's possible to set the cache timeout per method with z_cache / per call to cache_set,
        ...     # it's helpful to adjust default_cache_time to a number of seconds which is suitable for most
        ...     # methods in your class, avoiding the need to specify it each time..
        ...     default_cache_time: Union[float, int] = 300
        ...     # default_cache_key_time should generally be set to the same number of seconds as default_cache_time.
        ...     # It controls how long the "cache key log" is held for, which is simply a list of known cache keys for
        ...     # your class, enabling the use of the method .clear_all_cache_keys
        ...     default_cache_key_time: Union[float, int] = 300
        ...
        ...     @z_cache()
        ...     def hello(self, world=10):
        ...         sleep(5)
        ...         return f"Hello world: {world}"
        ...
        >>> c = MyClass()
        >>> # The first call to hello() will take 5 seconds due to the sleep
        >>> c.hello()
        Hello world: 10
        >>> c.hello()   # But when we call it again - it returns instantly
        Hello world: 10
        >>> # If we call .hello() with a different set of arguments, it will result in a new cache key being auto-generated,
        >>> # requiring a new, non-cached call to hello() to get the result for '5'
        >>> c.hello(5)
        Hello world: 5
    
    
    :param CACHE_MGR cls: For functions or methods which don't have a :class:`.CacheManagerMixin` based class or instance being passed
                          as their first argument, you'll need to set ``cls`` to point to a :class:`.CacheManagerMixin` based class
                          or instance, so that :func:`.z_cache` is able to both retrieve your overridden cache settings, and call
                          the various helper classmethod's for handling prefixed cache keys etc.
     
    :param str|callable cache_key: If you don't like automatic determination and argument-based generation of the cache_key, you
                                   may manually set the cache key to store the output of the method into.
                                   
                                   For dynamic generation of the cache key, this may be set to a ``callable``, such as a lambda, function,
                                   or method which returns the cache key to be used as a :class:`.str`.
                                   However, whatever callable is passed, must accept the exact same positional and keyword arguments as
                                   the method being wrapped - as those will be the args/kwargs passed to your ``cache_key``
                                   callable object.
    
    :param float|int cache_time:   The number of seconds to cache the output of the wrapped method/function for. By default, this is
                                   automatically extracted from the class of the wrapped function, by retrieving ``default_cache_time``
     
    :param format_args:            See the docs for :func:`.r_cache`
    :param format_opt:             See the docs for :func:`.r_cache`
    
    :param bool extract_class:     (Default: ``True``) This argument controls whether or not we attempt to auto-extract the
                                   related class/instance of the wrapped method by analyzing the first positional argument, along
                                   with the keyword arguments ``self`` and ``cls`` if they're present.
                                   
                                   If ``extract_class`` is set to ``False``, no attempt will be made to retrieve the related
                                   class/instance using the first argument - instead, if ``cls`` isn't passed, it will immediately
                                   fall back to using the original parent class for settings - :class:`.CacheManagerMixin`
    
    :param opts:                   See the docs for :func:`.r_cache`
    :return:
    """
    f_gen_cache_key = opts.get('gen_cache_key', cls.gen_cache_key if cls is not None else None)
    f_key_add_prefix = opts.get('key_add_prefix', cls.key_add_prefix if cls is not None else None)
    
    def _mk_key(_func: callable, _ck: Union[str, callable] = None, g_cachekey: callable = None, *zargs, **zkwargs):
        zargs = list(zargs)
        if _ck in [None, '', b'', False, True, 0, [], ()]:
            log.debug(" [_mk_key] No cache key specified, generating cache key based on function name + arguments")
            if len(zargs) > 0 and (inspect.isclass(zargs[0]) or inspect.ismethod(zargs[0])):
                log.debug(" [_mk_key] First argument appears to be cls/self - popping argument 0 to prevent wasteful cache repr() caching.")
                zargs.pop(0)
            gk = g_cachekey(*zargs, _comps_start=_func.__name__, **zkwargs)
            log.debug(" [_mk_key] Generated cache key: %s", gk)
            return gk
        return _ck
    
    # Sharing normal variables between the outer z_cache, and the intter wrapper function causes a lot of problems,
    # however, we're able to share an actual object instance easily - so we use a DictObject to make various variables
    # available from within the inner wrapper functions.
    z_opts = DictObject(
        {
            'cache_key':     cache_key, 'cache_time': cache_time, 'format_args': format_args, 'format_opt': format_opt,
            'extract_class': extract_class, 'gen_cache_key': f_gen_cache_key, 'key_add_prefix': f_key_add_prefix,
            'opts':          dict(opts)
        }
    )
    
    # _cls = cls
    
    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            kwargs = dict(kwargs)
            z_kwargs = dict(kwargs)
            zo = DictObject(z_opts)
            ex_class = zo.pop('extract_class', None)
            cset: Optional[CacheSettings] = None
            # gen_cache_key = zo.pop('gen_cache_key', None)
            # key_add_prefix = zo.pop('key_add_prefix', None)
            
            if ex_class:
                # First check if the first positional argument is a class, or an instance of CacheManagerMixin
                # If it is, and we don't already have a valid gen_cache_key set, then try to use gen_cache_key from that.
                if len(args) > 0:
                    if args[0] and (inspect.isclass(args[0]) or isinstance(args[0], CacheManagerMixin)):
                        log.debug(" [z_cache] Setting gen_cache_key to 'args[0].gen_cache_key'")
                        _cls: CacheManagerMixin = args[0]
                        if cset is None:
                            try:
                                cset = CacheSettings.from_class(_cls)
                            except Exception as e:
                                log.warning(f"Failed to generate CacheSettings object from class object: {_cls} "
                                            f"- reason: {type(e)} {str(e)}")
                            
                        # key_add_prefix = getattr(args[0], 'key_add_prefix', None) if key_add_prefix is None else key_add_prefix
                        # gen_cache_key = getattr(args[0], 'gen_cache_key', None) if gen_cache_key is None else gen_cache_key
                        #
                # If gen_cache_key is still None, see if 'self' or 'cls' are present in the keyword arguments
                # Just like with the positional args, we try to extract 'gen_cache_key' from the class/instance if it's possibly.
                if len(kwargs) > 0:
                    c_self, c_cls = kwargs.get('self'), kwargs.get('cls')
                    if c_self and (inspect.isclass(c_self) or isinstance(c_self, CacheManagerMixin)):
                        log.debug(" [z_cache] Setting gen_cache_key to 'c_self.gen_cache_key' (kwargs['self'])")
                        if cset is None:
                            try:
                                cset = CacheSettings.from_class(c_self)
                            except Exception as e:
                                log.warning(f"Failed to generate CacheSettings object from class object: {c_self} "
                                            f"- reason: {type(e)} {str(e)}")
                        # key_add_prefix = getattr(c_self, 'key_add_prefix', None) if key_add_prefix is None else key_add_prefix
                        # gen_cache_key = getattr(c_self, 'gen_cache_key', None) if gen_cache_key is None else gen_cache_key
                    if c_cls and (inspect.isclass(c_cls) or isinstance(c_cls, CacheManagerMixin)):
                        log.debug(" [z_cache] Setting gen_cache_key to 'c_cls.gen_cache_key' (kwargs['cls'])")
                        try:
                            cset = CacheSettings.from_class(c_cls)
                        except Exception as e:
                            log.warning(f"Failed to generate CacheSettings object from class object: {c_cls} "
                                        f"- reason: {type(e)} {str(e)}")
                        # key_add_prefix = getattr(c_cls, 'key_add_prefix', None) if key_add_prefix is None else key_add_prefix
                        # gen_cache_key = getattr(c_cls, 'gen_cache_key', None) if gen_cache_key is None else gen_cache_key
            
            cset: CacheSettings = CacheSettings.from_class(CacheManagerMixin) if cset is None else cset
            # zo = DictObject(z_opts)
            # If gen_cache_key is STILL None, then we fallback to CacheManagerMixin.gen_cache_key
            # gen_cache_key = CacheManagerMixin.gen_cache_key if gen_cache_key is None else gen_cache_key
            # If key_add_prefix is STILL None, then we fallback to CacheManagerMixin.key_add_prefix
            # key_add_prefix = CacheManagerMixin.key_add_prefix if key_add_prefix is None else key_add_prefix
            
            # If the decorator was passed a custom cache_key, then we need to pass it to key_add_prefix to ensure that the
            # key is both prefixed, and added to the known cache key list for easy removal later on.
            if zo.get('cache_key') is not None:
                log.debug(f" [z_cache] Calling key_add_prefix with cache key '{zo.cache_key}' + args ({args}) + kwargs ({kwargs})")
                # noinspection PyTypeChecker
                zo.cache_key = cset.key_add_prefix(zo.cache_key, call_args=list(args), call_kwargs=dict(kwargs))
                log.debug(f" [z_cache] New prefixed cache key is: '{zo.cache_key}'")
            # We use _mk_key - a wrapper for gen_cache_key, to auto-generate a cache key, depending on if the
            # passed cache_key was empty.
            zo.cache_key = _mk_key(f, zo.get('cache_key'), cset.gen_cache_key, *args, **z_kwargs)
            
            x_ck, x_ct = zo.pop('cache_key'), zo.pop('cache_time', None)
            x_ct = cset.default_cache_time if x_ct in [None, AUTO] else x_ct
            x_fmtargs, x_fmtopts = zo.pop('format_args', None), zo.pop('format_opt', None)
            z_kwargs.pop('_func', None), z_kwargs.pop('_ck', None), z_kwargs.pop('g_cachekey', None)
            log.debug(f" [z_cache] Calling r_cache: {x_ck}, {x_ct}, {x_fmtargs}, {x_fmtopts}, {z_opts.opts}")
            log.debug(f" [z_cache] func args: {args} || kwargs: {kwargs}")
            return r_cache(x_ck, x_ct, x_fmtargs, x_fmtopts, **zo.opts)(f)(*args, **kwargs)
        
        return wrapper
    
    return _decorator


# CacheSettings = namedtuple(
#     'CacheSettings', 'cache_prefix cache_key_log_name cache_sep default_cache_key_time default_cache_time'
# )


class CacheManagerMixin:
    """
    **CacheManagerMixin** is a class mixin which adds various methods and settings to assist with class-scoped caching, including
    an adjustable class-level cache prefix :attr:`.cache_prefix`, a method to remove all known cache keys managed by
    your class :meth:`.clear_all_cache_keys`, and a special decorator which automatically integrates with your class :func:`.z_cache`
    by intercepting the first argument of a method call.
    
    After you've extended :class:`.CacheManagerMixin`, you can enable argument-based caching for any method, simply by adding
    the decorator line ``@z_cache()`` above it. :func:`.z_cache` will automatically handle the cache key name, including extracting
    your method's passed arguments and inserting them into the cache key, ensuring method calls with certain arguments, are cached
    separately from other method calls with a different set of arguments.
    
    Most methods of this class are generally only used internally, or for certain niche use cases, but there are some methods
    that can be useful in almost any use case:
    
        * :meth:`.get_all_cache_keys` - returns a :class:`.set` of all known cache keys used by your class. Note that
          this method relies on the "cache key log", which will only be aware of keys that were cached within the
          last :attr:`.default_cache_key_time` seconds.
        
        * :meth:`.clear_cach_keys` - manually remove individual cache keys immediately
        
        * :meth:`.clear_all_cache_keys` - as the name implies, this method will remove **all known cache keys** that are found
          inside of the cache key log ( :attr:`.cache_key_log_name` )
        
        * :meth:`.cache_set` - set a cache key directly from the body of a method. The key that you enter, will be automatically
          prefixed with the appropriate :attr:`.cache_prefix`, so you don't need to worry about manually prefixing your cache key names.
        
        * :meth:`.cache_get` - retrieve the contents of a given cache key name - if it exists. Just like with :meth:`.cache_set`, this
          method will automatically prefix the key you enter. Additionally, by default, if the key you're trying to retrieve
          doesn't exist, it will make a second cache call without the prefix, in-case that you were really wanting the literal key
          you entered (not the auto-prefixed key). The automatic no-prefix-fallback can be disabled, see the docs for that method.
        
        * :meth:`.cache_get_or_set` - a combination of :meth:`.get` and :meth:`.set` - if the key you're trying to retrieve doesn't exist,
          then the fallback ``value`` will be set on that cache key and returned. Additionally, you may specify a "callback" for
          ``value``, i.e. a function/method which will be called if ``key`` doesn't exist, which should return the new value which
          should be set on the key and returned.
    
     
    Example::
        
        >>> from privex.helpers.cache.extras import CacheManagerMixin, z_cache
        >>> from time import sleep
        >>>
        >>> class MyClass(CacheManagerMixin):
        ...     # It's recommended, but not necessary - to re-create cache_key_lock with None, as it will ensure your
        ...     # class has a separate threading.Lock() instance from other CacheManagerMixin based classes
        ...     cache_key_lock: Optional[threading.Lock] = None
        ...     # You should override cache_prefix to ensure cache keys auto-created for your class won't
        ...     # conflict with cache keys created by other classes
        ...     cache_prefix: str = 'my_class'
        ...     # While it's possible to set the cache timeout per method with z_cache / per call to cache_set,
        ...     # it's helpful to adjust default_cache_time to a number of seconds which is suitable for most
        ...     # methods in your class, avoiding the need to specify it each time..
        ...     default_cache_time: Union[float, int] = 300
        ...     # default_cache_key_time should generally be set to the same number of seconds as default_cache_time.
        ...     # It controls how long the "cache key log" is held for, which is simply a list of known cache keys for
        ...     # your class, enabling the use of the method .clear_all_cache_keys
        ...     default_cache_key_time: Union[float, int] = 300
        ...
        ...     @z_cache()
        ...     def hello(self, world=10):
        ...         sleep(5)
        ...         return f"Hello world: {world}"
        ...
        >>> c = MyClass()
        >>> # The first call to hello() will take 5 seconds due to the sleep
        >>> c.hello()
        Hello world: 10
        >>> c.hello()   # But when we call it again - it returns instantly
        Hello world: 10
        >>> # If we call .hello() with a different set of arguments, it will result in a new cache key being auto-generated,
        >>> # requiring a new, non-cached call to hello() to get the result for '5'
        >>> c.hello(5)
        Hello world: 5
    
        
        
    """
    cache_key_lock: Optional[threading.Lock] = None
    """
    A :class:`threading.Lock` lock object which is used when adding cache keys to :attr:`.cache_key_log_name` to prevent
    the risk of two simultaneous cache key additions causing one call to overwrite the other's addition attempt.
    """
    cache_prefix: str = 'pvx_cmgr'
    """The prefix used for all cache keys generated for your class"""
    cache_key_log_name: str = 'all_cache_keys'
    """The cache key name component for the key used to store a list of all known cache keys for your class"""
    cache_sep: str = ':'
    """Separator character(s) used between identifying components within a cache key"""
    default_cache_key_time: Union[float, int] = 300
    """Default number of seconds to cache the log of known cache keys"""
    default_cache_time: Union[float, int] = 300
    """Default number of seconds to cache any objects"""
    
    @classmethod
    def _get_lock(cls, lock: ANY_LCK = None, timeout=30, fail=True) -> ContextManager:
        log.debug("Lock requested by caller: %s", caller_name())
        if lock == NO_LOCK:
            log.debug("Returning fake lock to caller: %s", caller_name())
            return fake_lock_manager()
        if lock is None:
            if cls.cache_key_lock is None:
                log.debug("Setting cls.cache_key_lock")
                cls.cache_key_lock = threading.Lock()
            log.debug("Returning cls.cache_key_lock")
            lock = cls.cache_key_lock
        log.debug("Returning lock_acquire_timeout(%s, timeout=%s, fail=%s)", lock, timeout, fail)
        return lock_acquire_timeout(lock, timeout=timeout, fail=fail)
    
    @classmethod
    def gen_cache_key(cls, *args, _auto_cache=True, _comps_start: list = None, _comps_end: list = None, **query) -> str:
        _comps_start, _comps_end = auto_list(_comps_start), auto_list(_comps_end)
        kcomps = [cls.cache_prefix] + _comps_start
        args, query = list(args), dict(query)
        if len(args) > 0:  # If there are positional args, join them up with commas and add to kcomps
            kcomps.append(','.join([str(a) for a in args]))
        if len(query) > 0:  # if there are kwargs, convert them into a list of strings 'key=value' and merge with kcomps
            ql = [f'{k}={v}' for k, v in query.items()]
            ql.sort()
            kcomps += ql
        kcomps += _comps_end
        # convert kcomps into a string by joining all items with the configured cache key separator
        k = cls.cache_sep.join(kcomps)
        if _auto_cache:
            cls.log_cache_key(k)
        return k
    
    @classmethod
    def get_all_cache_keys(cls) -> Set[str]:
        """
        Retrieve the list of cache keys as a :class:`.set` from the cache key ``'query_cache_keys'`` which
        stores the list of ``query_hidden:xxx:xxx:xxx`` keys, allowing for easy clearing of those cache keys when needed.
        :return:
        """
        lk = cls.cache_sep.join([cls.cache_prefix, cls.cache_key_log_name])
        qk = cached.get_or_set(lk, set(), cls.default_cache_key_time)
        qk = set() if empty(qk, True, True) else qk
        if isinstance(qk, (list, tuple)): qk = set(qk)
        return qk
    
    @classmethod
    def log_cache_key(cls, key: str, _lock: ANY_LCK = None) -> Set[str]:
        """
        Add a cache key name to the cache key log :attr:`.cache_key_log_name`. This usually doesn't need to be called from outside
        of this class, since most methods which may add or edit a cache key should also insert/update the key into the cache key log.
        
        :param str key:  The key to add to the cache key log.
        :param ANY_LCK _lock:  You may optionally pass a :class:`.Lock` instance if needed, e.g. to prevent a conflict
                               where the calling function/method has already acquired the class-level lock :attr:`.cache_key_lock`
                               It can also be set to the dummy type :class:`.NO_LOCK` ``NO_LOCK`` to prevent using a lock.
        :return Set[str] cache_key_log: The cache key log after adding ``ckeys``
        """
        lk = cls.cache_sep.join([cls.cache_prefix, cls.cache_key_log_name])
        with cls._get_lock(_lock):
            log.debug(f"Adding cache key '{key}' to list of cached query keys: '{lk}'")
            keys = cls.get_all_cache_keys()
            keys.add(key)
            cached.set(lk, keys, cls.default_cache_key_time)
            log.debug(f"Successfully added '{key}' to '{lk}' and wrote back to cache. Current key list: {keys}")
        return keys
    
    @classmethod
    def log_delete_cache_keys(cls, *ckeys: str, _lock: ANY_LCK = None) -> Set[str]:
        """
        Remove one or more cache key names from the cache key log :attr:`.cache_key_log_name`. This usually doesn't need to be called
        from outside of this class, since :meth:`.clear_cache_keys` automatically removes any logged cache key names after deleting
        the cache key itself from the global cache.
        
        :param str ckeys:      One or more cache keys to remove from the cache key log
        :param ANY_LCK _lock:  You may optionally pass a :class:`.Lock` instance if needed, e.g. to prevent a conflict
                               where the calling function/method has already acquired the class-level lock :attr:`.cache_key_lock`
                               It can also be set to the dummy type :class:`.NO_LOCK` ``NO_LOCK`` to prevent using a lock.
        :return Set[str] cache_key_log: The cache key log after removing ``ckeys``
        """
        lk = cls.cache_sep.join([cls.cache_prefix, cls.cache_key_log_name])
        ckeys = list(ckeys)
        log.debug(f"Removing {len(ckeys)} keys from cache key list. Keys to be removed: {ckeys}")
        log.debug(f" [log_delete_cache_keys] Obtaining hold on lock: {_lock}")
        with cls._get_lock(_lock):
            keys = cls.get_all_cache_keys()
            for key in ckeys:
                log.debug(f"Removing cache key '{key}' from list of cached query keys: '{lk}'")
                if key not in keys:
                    log.debug(f"The key '{key}' isn't in the key list. Skipping.")
                    continue
                keys.remove(key)
                log.debug(f"Successfully removed '{key}' from '{lk}'.")
            cached.set(lk, keys, cls.default_cache_key_time)
            log.debug(
                f"Finished removing {len(ckeys)} keys from cache key list. Wrote new key list back to cache. "
                f"Current key list: {keys}")
        log.debug(f" [log_delete_cache_keys] Releasing hold on lock: {_lock}")
        return keys
    
    log_delete_cache_key = log_delete_cache_keys
    
    @classmethod
    def cache_set(cls, key: str, value: Any, timeout: Optional[AnyNum] = AUTO, auto_prefix: bool = True):
        """
        This is a simple helper method which calls :meth:`.cached.set` - while automatically prepending :attr:`.cache_prefix`
        and :attr:`.cache_sep` before the key, plus when ``timeout=AUTO`` ( :class:`.AUTO` ), the timeout will be
        automatically set to the default timeout: :attr:`.default_cache_time`

        :param str key:   NOTE: Key will be auto-prepended with :attr:`.cache_prefixx - The cache key (as a string) to set the
                          value for, e.g. ``example:test``
        :param Any value: The value to store in the cache key ``key``
        :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
        :param bool auto_prefix: If set to ``True``, will auto-prepend :attr:`.cache_prefix` to ``key`` if it's not present.
        :return:
        """
        timeout = cls.default_cache_time if timeout == AUTO else timeout
        # For safety, if the cache key passed to this method already starts with `.cache_prefix`, then don't prepend
        # the cache_prefix to the key. Otherwise, if it's not present, then prepend the cache prefix.
        key = cls._pfx_key(key, auto_prefix=auto_prefix)
        return cached.set(key, value, timeout=timeout)
    
    @classmethod
    def cache_get(cls, key: str, default: Any = None, fail: bool = False, auto_prefix=True, fallback_prefix: bool = True):
        """
        This is a simple helper method which calls :meth:`.cached.get` - while automatically prepending :attr:`.cache_prefix`
        and :attr:`.cache_sep` before the key.

        :param str key: The cache key (as a string) to get the value for, e.g. ``example:test``
        :param Any default: If the cache key ``key`` isn't found / is expired, return this value (Default: ``None``)
        :param bool fail: If set to ``True``, will raise :class:`.CacheNotFound` instead of returning ``default``
                          when a key is non-existent / expired.
        :param bool auto_prefix: If set to ``True``, will auto-prepend :attr:`.cache_prefix` to ``key`` if it's not present.
        :param bool fallback_prefix: If set to ``True``, if we fail to find ``key`` with the prefix prepended, then we'll retry
                                     a cache lookup WITHOUT the key prefix.
        :raises CacheNotFound: Raised when ``fail=True`` and ``key`` was not found in cache / expired.
        :return Any value: The value of the cache key ``key``, or ``default`` if it wasn't found.
        """
        # timeout = cls.default_cache_time if timeout == AUTO else timeout
        # For safety, if the cache key passed to this method already starts with `.cache_prefix`, then don't prepend
        # the cache_prefix to the key. Otherwise, if it's not present, then prepend the cache prefix.
        key = cls._pfx_key(key, auto_prefix=auto_prefix)
        try:
            return cached.get(key, default, fail=True)
        except CacheNotFound as outer:
            # If fallback_prefix is enabled - try looking up 'key' with the cache prefix removed from the start of it
            if fallback_prefix and key.startswith(cls.cache_prefix):
                try:
                    key = cls._unpfx_key(key)  # Remove the prefix from the cache key
                    return cached.get(key, default, fail=True)
                except CacheNotFound as inner:
                    # If we still can't find the key at this point - give up and return the default value (or raise if fail=True)
                    if fail: raise inner
                    return default
            # If fallback_prefix isn't enabled, or the key wasn't prefixed - give up and raise the exception or return default value.
            if fail: raise outer
            return default
    
    @classmethod
    def cache_get_or_set(cls, key: str, value: Any, timeout: Optional[AnyNum] = AUTO, auto_prefix=True):
        """
        This is a simple helper method which calls :meth:`.cached.get_or_set` - while automatically prepending :attr:`.cache_prefix`
        and :attr:`.cache_sep` before the key.

        :param str key: The cache key (as a string) to get the value for, e.g. ``example:test``
        :param Any value: The value to store in the cache key ``key``
        :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
        :param bool auto_prefix: If set to ``True``, will auto-prepend :attr:`.cache_prefix` to ``key`` if it's not present.
        :raises CacheNotFound: Raised when ``fail=True`` and ``key`` was not found in cache / expired.
        :return Any value: The value of the cache key ``key``, or ``default`` if it wasn't found.
        """
        timeout = cls.default_cache_time if timeout == AUTO else timeout
        # For safety, if the cache key passed to this method already starts with `.cache_prefix`, then don't prepend
        # the cache_prefix to the key. Otherwise, if it's not present, then prepend the cache prefix.
        key = cls._pfx_key(key, auto_prefix=auto_prefix)
        return cached.get_or_set(key, value, timeout=timeout)
    
    @classmethod
    def _pfx_key(
            cls, key: Union[str, Callable[[Any], str]], auto_prefix: bool = True, _auto_cache=True,
            call_args: list = None, call_kwargs: dict = None, _lock: ANY_LCK = None
    ) -> str:
        """
        Add this class's cache key prefix to ``key`` if it isn't already prefixed
        
        
        :param str|callable key:   The key to prepend the cache key prefix onto - if not already prefixed.
                                   This may optionally be a function (e.g. a lambda) which returns a cache key to be auto-prefixed,
                                   and any necessary positional/keyword arguments for the function may be specified using the
                                   arguments ``call_args`` and ``call_kwargs``
        
        :param bool auto_prefix:   This argument is mainly used by internal methods to reduce the need to copy/paste handling code
                                   which allows users to request that a method does not attempt to auto-prefix the key they entered.
                                   
        :param bool _auto_cache:   This is a boolean key, which controls whether or not keys are automatically logged to the cache key log,
                                   which is a list of all known cache keys for a given class. Uses :meth:`.log_cache_key`
        
        :param list call_args:     If ``key`` is a callable (e.g. a lambda), ``call_args`` can be set to a list of positional arguments
                                   to pass to the callable function ``key``
                                   
        :param dict call_kwargs:   If ``key`` is a callable (e.g. a lambda), ``call_kwargs`` can be set to a :class:`.dict` of
                                   keyword arguments to pass to the callable function ``key``
        
        :param ANY_LCK _lock:      This method itself does not use a lock, but it calls upon :meth:`.log_cache_key` which does use
                                   a lock. You may optionally pass a :class:`.Lock` instance if needed, e.g. to prevent a conflict
                                   where the calling function/method has already acquired the class-level lock.
                                   It can also be set to the dummy type :class:`.NO_LOCK` ``NO_LOCK`` to prevent using a lock.
        :return str new_key:       The original ``key`` after it may or may not have had a prefix prepended to it.
        """
        if callable(key):
            call_args, call_kwargs = auto_list(call_args), empty_if(call_kwargs, {}, itr=True, zero=True)
            key = key(*call_args, **call_kwargs)
        key = key if key.startswith(cls.cache_prefix) or not auto_prefix else cls.cache_sep.join([cls.cache_prefix, key])
        if _auto_cache:   # By default, _auto_cache is enabled, which means we log all created keys to the cache key log
            cls.log_cache_key(key, _lock=_lock)
        return key
    
    key_add_prefix = _pfx_key
    
    @classmethod
    def _unpfx_key(cls, key: str, prefix: str = None, sep: str = None):
        """Remove the cache key prefix from a given cache key"""
        prefix, sep = empty_if(prefix, cls.cache_prefix), empty_if(sep, cls.cache_sep)
        j = key.split(f"{prefix}{sep}")
        return j[0] if len(j) == 1 else ''.join(j[1:])
    
    key_remove_prefix = _unpfx_key
    
    @classmethod
    def clear_cache_keys(cls, *keys: str, auto_prefix=True, remove_log=True, _lock: ANY_LCK = None) -> bool:
        log.debug(f" [clear_cache_keys] Obtaining hold on lock: {_lock}")
        with cls._get_lock(_lock):
            pk = [cls._pfx_key(k, auto_prefix=auto_prefix, _auto_cache=False) for k in keys]
            log.debug(f" [clear_cache_keys] Clearing {len(pk)} keys from cache: {pk}")
            res = cached.remove(*pk)
            if remove_log:
                log.debug(f" [clear_cache_keys] remove_log is True. Removing {len(pk)} keys from cache key log: {pk}")
                # nlock = threading.Lock()
                cls.log_delete_cache_keys(*pk, _lock=NO_LOCK)
        log.debug(f" [clear_cache_keys] Released hold on lock: {_lock}")
        
        return res
    
    @classmethod
    def clear_all_cache_keys(cls):
        """
        Remove all known cache keys related to this class which aren't expired.
        
        Uses the ``cache key log`` under the class specific key name :attr:`.cache_key_log_name`, which is a cached list
        that contains all known cache keys that have been created by this class, whether via :meth:`.cache_set`, or
        using the CacheManager decorator :func:`.z_cache`
        """
        keys = list(cls.get_all_cache_keys())
        if len(keys) < 1:
            return "NO_CACHE_KEYS"
        return cls.clear_cache_keys(*list(cls.get_all_cache_keys()))


